/*
 * SirvNET Services are copyright (c) 1998-2006 Trevor Klingbeil.
 *    http://www.sirv.net  ::  ftp.sirv.net
 * This program is copyrighted software; see the file COPYING for details.
 */

#include "services.h"
#include "log.h"
#include "timeout.h"


/************************************************************************************/
/************************************************************************************/

#define BLOCK_AUTH_EMAIL    1
#define GENERAL_BLOCK       2
#define BLOCK_SENDPASS      3

#define HELD_ID       1073741824
#define FLAG_HOLD     1
#define FLAG_MARK     2
#define FLAG_FREEZE   3

static void do_caccess(const char *source, const char *list, int level);

static inline void do_av(const char *source) { do_caccess(source, "AVOICE", 3); }
static inline void do_aop(const char *source) { do_caccess(source, "AOP", 5); }
static inline void do_sop(const char *source) { do_caccess(source, "SOP", 10); }
static inline void do_cf(const char *source) { do_caccess(source, "CFOUNDER", 13); }
#ifdef HAS_HALFOP
static inline void do_hop(const char *source) { do_caccess(source, "HOP", 4); }
#endif

static void help(const char *source);
static void ohelp(const char *source);
static void do_register(const char *source);
static void do_identify(const char *source);
static void do_akick(const char *source);
static void do_clear(const char *source, int clear);
static void do_clear2(const char *source) { do_clear(source, 0); }  /* user clear */
static void do_clear3(const char *source) { do_clear(source, 1); }  /* oper clear (expel) */
static void do_close(const char *source);
static void do_count(const char *source);
static void do_delete(const char *source);
static void do_deop(const char *source);
static void do_drop(const char *source);
static void do_flags(const char *source, int flags);
static void do_flist(const char *source);
static void do_freeze(const char *source) { do_flags(source, FLAG_FREEZE); }
static void do_getpass(const char *source);
static void do_hold(const char *source) { do_flags(source, FLAG_HOLD); }
static void do_info(const char *source);
static void do_invite(const char *source);
static void do_level(const char *source);
static void do_list(const char *source);
static void do_mark(const char *source) { do_flags(source, FLAG_MARK); }
static void do_official(const char *source);
static void do_op(const char *source);
static void do_remove(const char *source);
static void do_sendpass(const char *source);
static void do_set(const char *source);
static long do_set_ident(User *u, ChanInfo *ci, char *param);
static void do_set_desc(User *u, ChanInfo *ci, char *param);
static void do_set_founder(User *u, ChanInfo *ci, char *param);
static long do_set_keeptopic(User *u, ChanInfo *ci, char *param);
static long do_set_leaveops(User *u, ChanInfo *ci, char *param);
static void do_set_memo(User *u, ChanInfo *ci, char *param);
static void do_set_mlock(User *u, ChanInfo *ci, char *param);
static long do_set_notification(User *u, ChanInfo *ci, char *param);
static long do_set_opguard(User *u, ChanInfo *ci, char *param);
static void do_set_password(User *u, ChanInfo *ci, char *param);
static long do_set_private(User *u, ChanInfo *ci, char *param);
static long do_set_protect(User *u, ChanInfo *ci, char *param);
static long do_set_restricted(User *u, ChanInfo *ci, char *param);
static void do_set_successor(User *u, ChanInfo *ci, char *param);
static void do_set_topiclock(User *u, ChanInfo *ci, char *param);
static long do_set_trimban(User *u, ChanInfo *ci, char *param);
static long do_set_unsecure(User *u, ChanInfo *ci, char *param);
static void do_set_url(User *u, ChanInfo *ci, char *param);
static void do_set_welcome(User *u, ChanInfo *ci, char *param);
static void do_force_topic(const char *source);
static void do_unban(const char *source);
static void do_when(const char *source);
static void do_why(const char *source);

static int update_caccess(ChanInfo *ci, int level, const char *entry, const char *instater, const char *reason,
									int noop, int acc, unsigned long nickid);
static void mass_check_kick(const char *chan);


/************************************************************************************/

typedef struct {
	const char *cmd;
	void (*func)(const char *source);
	unsigned int access;
	unsigned int readonly;
	const char *desc;
} Command;

static Command commands[] = {
	{ "AKICK",	do_akick,	BASIC,	1,	CS_OHELP_AKICK },
	{ "AOP",	do_aop,		BASIC,	1,	CS_OHELP_AOP },
	{ "AVOICE",	do_av,		BASIC,	1,	CS_OHELP_AVOICE },
	{ "CF",		do_cf,		BASIC,	1,	CS_OHELP_CF },
	{ "CFOUNDER",	do_cf,		BASIC,	1,	CS_OHELP_CFOUNDER },
	{ "CLEAR",	do_clear2,	BASIC,	0,	CS_OHELP_CLEAR },
	{ "CLOSE",	do_close,	SA,	0,	CS_OHELP_CLOSE },
	{ "COUNT",	do_count,	BASIC,	0,	CS_OHELP_COUNT },
	{ "DELETE",	do_delete,	SRA,	1,	CS_OHELP_DELETE },
	{ "DEOP",	do_deop,	BASIC,	1,	CS_OHELP_DEOP },
	{ "DROP",	do_drop,	BASIC,	1,	CS_OHELP_DROP },
	{ "EXPEL",	do_clear3,	SOP,	0,	CS_OHELP_EXPEL },
	{ "FLIST",	do_flist,	OPER,	0,	CS_OHELP_FLIST },
	{ "FREEZE",	do_freeze,	SOP,	0,	CS_OHELP_FREEZE },
	{ "GETPASS",	do_getpass,	SA,	0,	CS_OHELP_GETPASS },
	{ "HELP",	help,		BASIC,	0,	CS_OHELP_HELP },
	{ "HOLD",	do_hold,	SA,	0,	CS_OHELP_HOLD },
#ifdef HAS_HALFOP
	{ "HOP",	do_hop,		BASIC,	0,	CS_OHELP_HOP },
#endif
	{ "IDENTIFY",	do_identify,	BASIC,	0,	CS_OHELP_IDENTIFY },
	{ "ID",		do_identify,	BASIC,	0,	CS_OHELP_ID },
	{ "INFO",	do_info,	BASIC,	0,	CS_OHELP_INFO },
	{ "INVITE",	do_invite,	BASIC,	0,	CS_OHELP_INVITE },
	{ "LEVEL",	do_level,	SOP,	0,	CS_OHELP_LEVEL },
	{ "LIST",	do_list,	BASIC,	0,	CS_OHELP_LIST },
	{ "MARK",	do_mark,	SA,	0,	CS_OHELP_MARK },
	{ "OHELP",	ohelp,		OPER,	0,	CS_OHELP_OHELP },
	{ "OFFICIAL",	do_official,	SRA,	0,	CS_OHELP_OFFICIAL },
	{ "OP",		do_op,		BASIC,	0,	CS_OHELP_OP },
	{ "REGISTER",	do_register,	BASIC,	1,	CS_OHELP_REGISTER },
	{ "REMOVE",	do_remove,	BASIC,	1,	CS_OHELP_REMOVE },
	{ "SENDPASS",	do_sendpass,	BASIC,	1,	CS_OHELP_SENDPASS },
	{ "SET",	do_set,		BASIC,	1,	CS_OHELP_SET },
	{ "TOPIC",	do_force_topic,	SOP,	1,	CS_OHELP_TOPIC },
	{ "UNBAN",	do_unban,	BASIC,	0,	CS_OHELP_UNBAN },
	{ "WHEN",	do_when,	BASIC,	0,	CS_OHELP_WHEN },
	{ "WHY",	do_why,		BASIC,	0,	CS_OHELP_WHY },
	{ NULL }
};

/************************************************************************************/

int check_cs_command(const char *cmd) {
	Command *c;
	int i=0;

	for (c = commands; c->cmd; c++) {
		if (sstrcmp(cmd, c->cmd) == 0) {
			return i;
		}
		i++;
	}
	return -1;
}

/************************************************************************************/

inline void update_cs_table(int cmd, int newlevel) {
	commands[cmd].access = newlevel;
}

/************************************************************************************/

static void timeout_clear(Timeout *t)
{
	ChanInfo *ci = findchaninfo(t->data);

	if (ci)
		ci->flags &= ~CS_SUSPEND;

	send_cmd(s_ChanServ, "MODE %s -lb *!*@*", t->data);
	send_cmd(s_ChanServ, "PART %s", t->data);
	free(t->data);
}

/************************************************************************************/

static void remove_ban(Timeout *t)
{
	char *s = sstrdup(t->data);
	char *chan = strtok(s, " ");
	char *mask = strtok(NULL, "");

	send_cmd(s_ChanServ, "MODE %s -b %s", chan, mask);
	send_cmd(s_ChanServ, "PART %s", chan);

	free(s);
	free(t->data);
}


/************************************************************************************/

static int match_host(User *u, char *host)
{
	char *buf;
	int i;

	if (!u || !host)
		return 0;

	buf = smalloc(strlen(u->nick) + strlen(u->username) + strlen(u->host) + 3);
	sprintf(buf, "%s!%s@%s", u->nick, u->username, u->host);


	i=match_wild_nocase(host, buf);

	free(buf);

	return i ? 1 : 0;

}

/************************************************************************************/

static int check_maxchans(int id)
{
	char buf[256];
	int i=0;

        snprintf(buf, sizeof(buf), "SELECT name FROM chan WHERE fid = %d", id);

        if (!mysql_query(&mysql, buf)) {
                if ((res2 = mysql_store_result(&mysql)) != NULL) {
                        i = mysql_num_rows(res2);
			mysql_free_result(res2);
			return i;
		}
        }

	return 0;
}

/************************************************************************************/

void check_welcome(const char *source, const char *chan)
{
	ChanInfo *ci = findchaninfo(chan);


	if (cburst && ci && ci->welcome && !is_frozen(chan))
		notice(s_ChanServ, source,
			"%s Welcome: %s", chan, ci->welcome);

	if (cburst && ci && ci->flags & CS_OFFICIAL)
		notice(s_ChanServ, source,
			"You have joined an official network channel of %s (%s).",
				NETWORK_NAME, chan);

}

/************************************************************************************/

static void add_id_chan(User *u, ChanInfo *ci, int reg)
{
	struct u_founderlist *fl;


	if (!ci)
		return;


	for (fl = u->fchans; fl; fl = fl->next) {
		if (fl->chan == ci)
			return;
	}


	if (!reg && LOG_IDENTIFIES)
		blog("CS_I", "[CS] I %s (%s!%s@%s)",
			ci->name, u->nick, u->username, u->host);

	fl = scalloc(sizeof(struct u_founderlist), 1);
	fl->next = u->fchans;
	fl->prev = NULL;
	if (u->fchans)
		u->fchans->prev = fl;
	u->fchans = fl;
	fl->chan = ci;
}



/************************************************************************************/

static inline void add_idchan(User *u, ChanInfo *ci)
{
	add_id_chan(u, ci, 0);
}

/************************************************************************************/

static int is_founder(User *u, const char *chan)
{
	ChanInfo *ci = findchaninfo(chan);
	struct u_founderlist *fl;

	if (!u || !ci)
		return 0;

	for (fl = u->fchans; fl; fl = fl->next) {
		if (fl->chan == ci)
			return 1;
	}

       /* We do a strcmp to make sure the user is using the founder's nick at the time of the cmd. */

	if (ci->flags & CS_UNSECURE) {
		if (is_recognized(u->nick, ci->founder) && !sstrcmp(u->nick, ci->founder))
			return 2;
	}

	return 0;
}

/************************************************************************************/

static inline void nick_trace(char **ntrace, char *ptr)
{
	if (ntrace)
		*ntrace=ptr;
}


/************************************************************************************/

static int match_nick_ban(User *u, const char *bannick)
{
	NickInfo *ni = findnick(bannick);
	char buf[256], mbuf[HOSTMAX];
	unsigned long nickid;

	if (!ni)
		return 0;

	nickid = ni->id;

	snprintf(buf, sizeof(buf), "SELECT usermask FROM nick WHERE id=%ld", nickid);
	if (!mysql_query(&mysql, buf)) {
		if ((res2 = mysql_store_result(&mysql)) != NULL) {
			if (mysql_num_rows(res2)) {
				row2 = fetch(res2);
				snprintf(mbuf, sizeof(mbuf), "*!%s", row2[0]);
				if (match_host(u, mbuf)) {
					mysql_free_result(res2);
					return 1;
				}
			}
			mysql_free_result(res2);
			return 0;
		}
	}
	return 0;
}

/************************************************************************************/

/* caccess paramater is to bypass founder access checking since we only want to give
 * users who have identified to the channel SET privileges and not access list manipulation.
 * Otherwise, anyone identified to the channel could make access list changes and we have no
 * way to trace what appropriate nick added the entry. Therefore, identifying to a channel only
 * gives SET permissions. To manipulate the access list, the user will need to be either identified
 * to the founder's nick, or be authenticated by a SOP+ entry. */


int get_access(User *u, const char *chan, char **ntrace, int caccess)
{
	ChanInfo *ci = findchaninfo(chan);
	struct AccessList *a, *top=NULL;
	struct u_idlist *nl;
	int facc=0;


	if (!ci || !u)
		return 0;



	if (!caccess && (facc=is_founder(u, chan))) {
		nick_trace(&*ntrace, ci->founder);
		return facc==1 ? 15 : 14;      /* 14 = Founder recognized w/ unsecure. */
	}

			/* IDENT IS OFF */



	if (!(ci->flags & CS_IDENT)) {
		if (is_recognized(u->nick, ci->founder)) {   /* does user's hostmask match the founder's access list? */
			nick_trace(&*ntrace, ci->founder);
			return 15;
		}

		for (a = ci->entries; a; a = a->next) {
			if (strchr(a->entry, '@')) {			  /* hostmask entry */
				if (match_host(u, a->entry)) {
					if (!top || (a->level > top->level))
						top = a;
				}

			} else {  					 /* nick entry */
				if (a->level == -2) {    /* nick akick entry */
					if (!top && match_nick_ban(u, a->entry))
						top = a;
				} else if (is_recognized(u->nick, a->entry)) {
					if (!is_frozen(a->entry) && (!top || (a->level > top->level)))
						top = a;
				}

			}
		}



			/* IDENT IS ON */



	} else if (ci->flags & CS_IDENT) {
		if (is_identified(u->nick, ci->founder)) {
			nick_trace(&*ntrace, ci->founder);
			return 15;
		}

		for (a = ci->entries; a; a = a->next) {
			if (strchr(a->entry, '@')) {			  /* akick host entry */
				if (match_host(u, a->entry)) {
					if (!top)
						top = a;
				}
			}
			if (a->level == -2) {    /* nick akick entry */
				if (!top && match_nick_ban(u, a->entry))
					top = a;
			}

			for (nl = u->nicks; nl; nl = nl->next) {
				if (sstrcmp(a->entry, nl->nick) == 0) {
					if (!top || (a->level > top->level))
						top = a;
				}
			}
		}
	}


	if (top) {
		nick_trace(&*ntrace, top->entry);
		return top->level;
	} else
		return 0;
}


/************************************************************************************/

void expire_chanlists(const char *nick)
{

	ChanInfo *ci;
	struct AccessList *a, *a2;
	char buf[512];
	int num_rows=0, i=0;

	snprintf(buf, sizeof(buf),
		"SELECT name FROM chan");

	if (!mysql_query(&mysql, buf)) {
		if ((res2 = mysql_store_result(&mysql)) != NULL)
			num_rows = mysql_num_rows(res2);
	}

	for (i=0; i < num_rows; i++) {
		row2 = fetch(res2);
		ci = findchaninfo(row2[0]);
		a = ci->entries;
		while (a) {
			a2 = a->next;
			if (sstrcmp(a->entry, nick) == 0) {
				User *u = finduser(ci->founder);
				NickInfo *ni = findnick(ci->founder);
				char buf[256 + NICKMAX + CHANMAX];

				if (ci->flags & CS_NOTIFY) {
					snprintf(buf, sizeof(buf), "Nick '%s' has expired and has been removed "
						"from the %s list of %s.",
						nick, a->level==3 ? "Auto-Voice" : a->level==5 ? "Auto-Op" :
						a->level==10 ? "Super-Op" : a->level==13 ? "Co-Founder" : "(Unknown)",
						ci->name);
					send_cmd(s_ChanServ, "NOTICE @%s :%s", ci->name, buf);
					send_sql("INSERT INTO memos VALUES "
						"('%s','%s',NULL,'%s',%d,1,0,0)",
						ci->founder, s_MemoServ, buf, time(NULL));
				}
				if (ni && ni->flags & MS_EMAILMEMOS)
					email_memo(ni, buf, s_MemoServ);
				if (u)
					check_memos(u, is_identified(u->nick, u->nick) ? 1 : 0);


				if (a->prev)
					a->prev->next = a->next;
				else
					ci->entries = a->next;
				if (a->next)
					a->next->prev = a->prev;
				if (a->reason)
					free(a->reason);
				free(a->entry);
				free(a->instater);
				free(a);
			}


			a = a2;
		}
	}

	if (num_rows)
		mysql_free_result(res2);
}

/************************************************************************************/

int check_topiclock(ChanInfo *ci, const char *source)
{
	User *u = finduser(source);


	if (!ci)
		return 0;

	if (!u)
		return 1;

	if (get_access(u, ci->name, NULL, 0) < ci->topiclock)
		return 1;

	return 0;

}

/************************************************************************************/

/*
  Return values:
     -5 = Access denied for this manipulation level.
     -4 = No-Op set, deny adding to list (permits manipulation of existing access).
     -3 = no change, access list at max capacity.
     -2 = no change, channel at maximum capacity for all access lists.
     -1 = no change, entry already on that list.
      0 = General failure.
      1 = Entry added to list.
      2 = Entry updated on access list.
      3 = Entry deleted on access list.
*/

static int update_caccess(ChanInfo *ci, int level, const char *entry, const char *instater, const char *reason, 
									int noop, int acc, unsigned long nickid)
{
	struct AccessList *a;
	int countall=0, countlist=0;

	if (level && !instater)
		return 0;


	for (a = ci->entries; a; a = a->next) {

		countall++;
		if (a->level == level)
			countlist++;

		if (sstrcmp(a->entry, entry) == 0) {
			if (!level) {  /* delete record */
				if (a->prev)
					a->prev->next = a->next;
				else
					ci->entries = a->next;
				if (a->next)
					a->next->prev = a->prev;
				send_sql("DELETE FROM caccess WHERE chan=%d AND entry='%s'",
					ci->id, entry);
				if (a->reason)
					free(a->reason);
				if (ci->flags & CS_NOTIFY) {
					User *u = finduser(ci->founder);
					NickInfo *ni = findnick(ci->founder);
					char buf[256 + (CHANMAX*2) + (NICKMAX*2)];
					snprintf(buf, sizeof(buf),
						"%s has deleted '%s' from the %s list of channel %s.",
						instater, a->entry,
						a->level<0 ? "AKICK" : a->level==3 ? "AVOICE" :
						a->level==5 ? "AOP" : a->level==10 ? "SOP" :
						a->level==13 ? "CFOUNDER" : "", ci->name);
					send_sql("INSERT INTO memos VALUES "
						"('%s','%s',NULL,'%s',%d,1,0,0)",
						ci->founder, s_MemoServ, buf, time(NULL));

 					send_cmd(s_ChanServ, "NOTICE @%s :%s", ci->name, buf);
					if (ni && ni->flags & MS_EMAILMEMOS)
						email_memo(ni, buf, s_MemoServ);
					if (u)
						check_memos(u, is_identified(u->nick, u->nick) ? 1 : 0);


				}
				free(a->entry); free(a->instater); free(a);
				return 3;
			}

			if (a->level == level)
				return -1;

			if (a->level >= acc)
				return -5;


			if (ci->flags & CS_NOTIFY) {
				User *u = finduser(ci->founder);
				NickInfo *ni = findnick(ci->founder);
				char buf[256 + CHANMAX + (NICKMAX*2) + HOSTMAX];
				snprintf(buf, sizeof(buf),
					"%s has changed '%s' from the %s list to the %s "
					"list of channel %s.",	instater, a->entry,
					a->level<0 ? "AKICK" : a->level==3 ? "AVOICE" :
					a->level==5 ? "AOP" : a->level==10 ? "SOP" :
					a->level==13 ? "CFOUNDER" : "",
					level==-1 ? "AKICK" : level==3 ? "AVOICE" :
					level==5 ? "AOP" : level==10 ? "SOP" :
					level==13 ? "CFOUNDER" : "", ci->name);
				send_sql("INSERT INTO memos VALUES "
					"('%s','%s',NULL,'%s',%d,1,0,0)",
					ci->founder, s_MemoServ, buf, time(NULL));
				send_cmd(s_ChanServ, "NOTICE @%s :%s", ci->name, buf);
				if (ni && ni->flags & MS_EMAILMEMOS)
					email_memo(ni, buf, s_MemoServ);
				if (u)
					check_memos(u, is_identified(u->nick, u->nick) ? 1 : 0);
			}


			a->level = level;
			if (a->instater)
				free(a->instater);
			a->instater = sstrdup(instater);
			a->when = time(NULL);
			a->reason = reason ? sstrdup(reason) : NULL;
			a->nickid = nickid;

			if (!reason)
				send_sql("UPDATE caccess SET level=%d, instater='%s', reason=NULL "
					"WHERE chan = %d AND entry='%s'", level, instater, ci->id, entry);
			else
				send_sql("UPDATE caccess SET level=%d, instater='%s', reason='%s' "
					"WHERE chan = %d AND entry='%s'",
						level, instater, reason, ci->id, entry);

			return 2;
		}
	}

	if (noop)
		return -4;


	if (CHAN_ACC_MAX && (countall > CHAN_ACC_MAX))
		return -2;

	if (CHAN_LIST_MAX && (countlist > CHAN_LIST_MAX))
		return -3;


	if (level==0)
		return 0;


	a = scalloc(sizeof(struct AccessList), 1);
	a->next = ci->entries;
	a->prev = NULL;
	if (ci->entries)
		ci->entries->prev = a;
	ci->entries = a;
	a->entry = sstrdup(entry);
	a->instater = sstrdup(instater);
	a->level = level;
	a->when = time(NULL);
	a->reason = reason ? sstrdup(reason) : NULL;
	a->nickid = nickid;


	if (ci->flags & CS_NOTIFY) {
		User *u = finduser(ci->founder);
		NickInfo *ni = findnick(ci->founder);
		char buf[256 + CHANMAX + NICKMAX + HOSTMAX];
		snprintf(buf, sizeof(buf),
			"%s has added '%s' to the %s list of channel %s.",
			instater, a->entry,
			a->level<0 ? "AKICK" : a->level==3 ? "AVOICE" :
			a->level==5 ? "AOP" : a->level==10 ? "SOP" :
			a->level==13 ? "CFOUNDER" : "", ci->name);

		send_sql("INSERT INTO memos VALUES "
			"('%s','%s',NULL,'%s',%d,1,0,0)",
			ci->founder, s_MemoServ, buf, time(NULL));
		send_cmd(s_ChanServ, "NOTICE @%s :%s", ci->name, buf);
		if (ni && ni->flags & MS_EMAILMEMOS)
			email_memo(ni, buf, s_MemoServ);
		if (u)
			check_memos(u, is_identified(u->nick, u->nick) ? 1 : 0);

	}

	if (!reason)
		send_sql("INSERT INTO caccess VALUES (%d,'%s',%d,'%s',%d,NULL)",
			ci->id, entry, level, instater, time(NULL));
	else
		send_sql("INSERT INTO caccess VALUES (%d,'%s',%d,'%s',%d,'%s')",
			ci->id, entry, level, instater, time(NULL), reason);
	return 1;

}


/************************************************************************************/

int check_op_protect(User *u, const char *chan)
{
	ChanInfo *ci = findchaninfo(chan);

	if (!u || !ci || (!(ci->flags & CS_PROTECT)))
		return 0;

	if (is_frozen(chan))
		return 0;


 /* check if !is_chanop() to make sure they are an op to begin with. Otherwise, if someone sets -o to a client
    that isn't opped, ChanServ will set them +o. */
  
	if (!is_chanop(u->nick, chan))
		return 0;

	if (get_access(u, chan, NULL, 0) >= 5) {  /* keep above 3 or 4 or voices and half-ops will be oped */
		send_cmd(s_ChanServ, "MODE %s +o %s", chan, u->nick);
		return 1;
	}

	return 0;
}

/************************************************************************************/


int check_valid_op(User *u, const char *chan)
{
	ChanInfo *ci = findchaninfo(chan);
	Channel *c = findchan(chan);
	int acc=0;

	if (!ci || !u || !c)
		return 1;


	acc = get_access(u, chan, NULL, 0);

	if (((ci->flags & CS_OPGUARD) && (acc < 5)) || is_frozen(chan))  {
		send_cmd(s_ChanServ, "MODE %s -o %s", chan, u->nick);
		return 0;
	}

	if (acc>0 && cburst)
		send_sql("UPDATE chan SET lasttime=%d WHERE id=%d", time(NULL), ci->id);


	if (!c->users->next && (ci->flags & CS_LEAVEOPS))
		return 1;


	if ((acc < 5) && !c->users->next && (!(ci->flags & CS_LEAVEOPS))) {
		send_cmd(s_ChanServ, "MODE %s -o %s", chan, u->nick);
		return 0;
	}


	return 1;
}

/************************************************************************************/

#ifdef UNREAL

void check_special_op(User *u, const char *chan)
{
	NickInfo *ni;
	ChanInfo *ci = findchaninfo(chan);
	int acc=0;

	if (!ci || !u)
		return;

	if (is_frozen(chan) || is_frozen(u->nick))
		return;

	if ((ni = findnick(u->nick)) && ni->flags & NS_NEVEROP)
		return;

	acc = get_access(u, chan, NULL, 0);

	if (acc==13)
		send_cmd(s_ChanServ, "MODE %s +a %s %s", chan, u->nick);
	else if (acc>13)
		send_cmd(s_ChanServ, "MODE %s +q %s %s", chan, u->nick);
}

#endif

/************************************************************************************/

int check_should_op(User *u, const char *chan)
{
	NickInfo *ni;
	ChanInfo *ci = findchaninfo(chan);
	int acc=0;

	if (!ci || !u)
		return 0;

	if (is_chanop(u->nick, chan))
		return 0;

	if (is_frozen(chan) || is_frozen(u->nick))
		return 0;


	if ((ni = findnick(u->nick)) && ni->flags & NS_NEVEROP)
		return 0;

	acc = get_access(u, chan, NULL, 0);


	if (acc > 3) {

#ifdef HAS_HALFOP
		if (acc==4)
			send_cmd(s_ChanServ, "MODE %s +h %s %s", chan, u->nick);
#endif
#ifdef UNREAL
		else if (acc==13)
			send_cmd(s_ChanServ, "MODE %s +ao %s %s", chan, u->nick, u->nick);
		else if (acc>13)
			send_cmd(s_ChanServ, "MODE %s +qo %s %s", chan, u->nick, u->nick);
#endif
#ifdef HAS_HALFOP
		else
			send_cmd(s_ChanServ, "MODE %s +o %s", chan, u->nick);
		if (cburst)
			send_sql("UPDATE chan SET lasttime=%d WHERE id=%d", time(NULL), ci->id);
		return (acc==4) ? 4 : 1;
#else /* !HAS_HALFOP */
		send_cmd(s_ChanServ, "MODE %s +o %s", chan, u->nick);
		if (cburst)
			send_sql("UPDATE chan SET lasttime=%d WHERE id=%d", time(NULL), ci->id);

		return 1;
#endif

	} else if (acc==3 && !is_voiced(u->nick, chan)) {
		send_cmd(s_ChanServ, "MODE %s +v %s", chan, u->nick);
		return 3;
	} else
		return 0;
}


/************************************************************************************/

static int level_access(ChanInfo *ci, char **caccess)
{
	if (ci->flags & CS_LEVELOPER) {
		*caccess = sstrdup("This channel is restricted to IRC Operators.");
		return 1;
	} else if (ci->flags & CS_LEVELSOP) {
		*caccess = sstrdup("This channel is restricted to services operators.");
		return 2;
	} else if (ci->flags & CS_LEVELSA) {
		*caccess = sstrdup("This channel is restricted to services administrators.");
		return 3;
	} else if (ci->flags & CS_LEVELSRA) {
		*caccess = sstrdup("This channel is restricted to services roots.");
		return 4;
	} else if (ci->flags & CS_LEVELMASTER) {
		*caccess = sstrdup("This channel is restricted to services root masters.");
		return 5;
	}
	return 0;
}

/************************************************************************************/

int check_should_kick(User *u, const char *chan)
{
	ChanInfo *ci = findchaninfo(chan);
	Channel *c = findchan(chan);
	char *access, *amask=0;
	int acc=0, closed=chan_closed(chan);
	unsigned int level=0;
	int isoper = (u->mode & UMODE_O);


	if (!closed) {
		if (!ci)
			return 0;


		else if (is_frozen(chan))
			return 0;


		acc = get_access(u, chan, &amask, 0);
	}
	if (closed || (acc < 0) || ((acc < 3)  && ci->flags & CS_RESTRICTED) || 
			(level=level_access(ci, &access)) > u->os_level+1) {
		char *reason = level ? access : "You are not permitted to be in this channel.";

		if (closed && !isoper) {
			Timeout *t = add_timeout(CHAN_INHABIT, timeout_clear, 0);
			t->data = sstrdup(chan);
			send_cmd(s_ChanServ, "MODE %s +lb 1 *!*@*", chan);
			send_cmd(s_ChanServ, "JOIN %s", chan);
		}

		if ((isoper && level) || (!isoper && !closed) ) {
			int freemask=0;
#ifdef SOLID
			if (u->maskhost)
				amask = u->maskhost;
#endif
			if (!amask || (acc == -2)) {  /* -2 for nick entry akicks */
				char *tmask = create_mask(u);
				amask = smalloc(strlen(tmask) + 3);
				sprintf(amask, "*!%s", tmask);
				free(tmask);
				freemask=1;
			}

#ifdef BAHAMUT
			send_cmd(s_ChanServ, "SVSMODE %s -e %s", chan, u->nick);
#endif
			if (c)
				chan_addban(chan, amask, 1);
			else {
				Timeout *t = add_timeout(CHAN_INHABIT, timeout_clear, 0);
				Timeout *t2 = add_timeout(CHAN_INHABIT, remove_ban, 0);
				char data[512];

				snprintf(data, sizeof(data), "%s %s", chan, amask);
				t2->data = sstrdup(data);
				t->data = sstrdup(chan);

				send_cmd(s_ChanServ, "JOIN %s", chan);
				send_cmd(s_ChanServ, "MODE %s +b :%s", chan, amask);
			}

			if (freemask)
				free(amask);
		}

		if (!isoper || (isoper && level)) {
			send_cmd(s_ChanServ, "KICK %s %s :%s", chan, u->nick, reason);
			if (level)
				free(access);
			return 1;
		} else
			return 0;
	}
	if (level)
		free(access);

	return 0;

}

/************************************************************************************/


void chanserv(const char *source, char *buf)
{
	Command *c;
	const char *cmd = gettok(buf, " ", CMDMAX);
	int valid=0;

	if (!cmd)
		return;

	if (!sql_alive) {
		notice(s_ChanServ, source, "Sorry, we're currently experiencing temporary "
			"technical difficulties. Please try again shortly.");
		return;
	}


/* We single the SOP command out or else it'll conflict with permissions for Oper Services' SOP command. */
	if (sstrcmp(cmd, "SOP") == 0) {
		do_sop(source);
		return;
	}

	for (c = commands; c->cmd; c++) {
		if (sstrcmp(c->cmd, cmd) == 0) {
			valid=1;
			if (READ_ONLY && c->readonly)
				notice(s_ChanServ, source,
					"Sorry, that command is temporarily disabled while services are in "
					"read-only mode.");
			else if (!c->access || has_access(source, c->access))
				c->func(source);
			else
				notice(s_ChanServ, source, "Permission denied.");
			break;
		}
	}
	if (!valid)
		notice(s_ChanServ, source,
			"Unrecognized command \2%s\2. Type \"/%s HELP\" for help.",
			cmd, s_ChanServ);
}

/************************************************************************************/

static void do_caccess(const char *source, const char *list, int level)
{
	ChanInfo *ci;
	const char *chan = gettok(NULL, " ", CHANMAX);
	const char *cmd = gettok(NULL, " ", CMDMAX);
	const char *param = strtok(NULL, " ");
	const char *reason = strtok(NULL, "");
	int mask=0, acc=0, acc2=0, isoper=0;
	char *ntrace;


	if ( (!chan || !cmd) ||
		((sstrcmp(cmd, "LIST") && sstrcmp(cmd, "WIPE") && sstrcmp(cmd, "ADD") && sstrcmp(cmd, "DEL")) ||
		((!sstrcmp(cmd, "ADD") || !sstrcmp(cmd, "DEL")) && !param))) {
		notice(s_ChanServ, source,
			"Syntax: /%s %s <#channel> [ADD|DEL|LIST|WIPE] [<nick>] [reason]", s_ChanServ, list);
		notice(s_ChanServ, source,
			"\2/%s HELP %s\2 for more information.",
			s_ChanServ, list);



	} else if (!(ci=findchaninfo(chan))) {
		notice(s_ChanServ, source,
			"%s ERROR: Channel %s is not registered!", list, chan);


	} else if ((acc=get_access(finduser(source), chan, &ntrace, 0)) < 3 && !(isoper=is_oper(source))) {
		notice(s_ChanServ, source,
			"%s ERROR: You do not have access for this command on channel %s.", list, chan);


	} else if (isoper && sstrcmp(cmd, "LIST") != 0 && acc < 5) {
		notice(s_ChanServ, source,
			"%s ERROR: You do not have access for this command on channel %s.", list, chan);


	} else if (sstrcmp(cmd, "LIST") == 0) {
		struct AccessList *a;
		int i=0, x=0, shown=0;
		char *lmask;

		for (a = ci->entries; a; a = a->next) {
			if (a->level == level) {
				lmask=0;
				if (!shown)
					notice(s_ChanServ, source,
						"%s list for \2%s\2:", list, chan);
				if ((x=sql_select("usermask FROM nick WHERE name = '%s'", a->entry))) {
					row = fetch(res);

					shown=1;
					if (MASK_HOSTNAME)
						lmask = mask_host(row[0]);

					notice(s_ChanServ, source,
						"  %d) %s   %s%s%s  [Added by: %s] %s%s%s", ++i, a->entry, 
						x ? "(" : "", x && lmask && !isoper ? lmask : row[0], x ? ")" : "", a->instater,
						a->reason ? "(" : "", a->reason ? a->reason : "", a->reason ? ")" : "");
					if (lmask)
						free(lmask);
				} else {   /* mask entry */
					notice(s_ChanServ, source,
						"  %d) %s   [Added by: %s] %s%s%s", 
						++i, a->entry, a->instater,
						a->reason ? "(" : "", a->reason ? a->reason : "", a->reason ? ")" : "");

				}
			}
		}
		if (!i)
			notice(s_ChanServ, source, "No %s entries on record for channel %s",list, chan);
		else
			notice(s_ChanServ, source, "\2End of %s List\2 (%d entries found)", list, i);



	} else if (acc <= level || (acc==5 && level==3)) {
		notice(s_ChanServ, source,
			"%s ERROR: You do not have access for this command on channel %s.", list, chan);



/* User is only identified to channel (SET permissions only) */
	} else if (acc==15 && (((acc2=get_access(finduser(source), chan, &ntrace, 1)) <= level)
		|| (acc2==5 && level==3))) {
		notice(s_ChanServ, source,
			"%s ERROR: You may not manipulate the %s list because you are not identified "
			"to any nick on the access list of %s. Identifying to a channel only gives you "
			"access to the SET commands.", list, list, chan); 


	} else if (sstrcmp(cmd, "ADD") == 0) {
		NickInfo *ni;
		const char *nick = param;
		int result=0;


		if (!(ni = findnick(nick))) {
			char *ptr = strchr(nick, '!');
			if (ptr)
				*ptr++;

			if (!strchr(nick, '!') && !strchr(nick, '@')) {
				notice(s_ChanServ, source,
					"%s ERROR: Nick '%s' is not registered!", list, nick);
				return;

			} else if (ci->flags & CS_IDENT) {
				notice(s_ChanServ, source,
					"%s ERROR: You may not add mask entries when %s has the IDENT "
					"option enabled. Only registered nicknames are permitted.",
					list, chan);
				return;

			} else if (strlen(nick) > 60) {
				notice(s_ChanServ, source,
					"%s ERROR: Length of mask is too long!", list);
				return;

			} else if (!strchr(nick, '!') || !strchr(nick, '@')) {
				notice(s_ChanServ, source,
					"%s ERROR: Mask entries must be in the form of nick!user@host.", list);
				return;


			} else if (*nick == '!' || *ptr == '@' || !check_valid_mask(ptr)) {
				notice(s_ChanServ, source,
				"%s ERROR: Invalid mask - Must be in the format of nick!user@host. "
				"(wildcards '*' are valid)", list);
				return;

			} else
				mask=1;

		}

		if (ni) {
			if (ni->flags < 0) {
				notice(s_ChanServ, source,
					"%s ERROR: Forbidden nick '%s' may not be added to channel access lists.",
					list, nick);
				return;
			}
			if (NSECURE && !is_authed(ni->nick)) {
				notice(s_ChanServ, source,
					"%s ERROR: Nick '%s' is not authorized and may not be added to channel "
					"access lists.", list, nick);
				return;
			}
		}

		result = update_caccess(ci, level, nick, ntrace, reason, ni ? ni->flags & NS_NOOP : 0, acc, 0);


		switch(result)
		{
			case -5:
				notice(s_ChanServ, source,
					"%s ERROR: You may not manipulate the access of %s on channel %s.",
					 list, nick, chan);
				return;

				break;
			case -4:
				notice(s_ChanServ, source,
					"Nick '%s' wishes not to be added to any channel access lists.",nick);
				return;

				break;
			case -3:
				notice(s_ChanServ, source,
					"%s ERROR: Channel %s is at it's maximum capicty for %s entries.",
						list, chan, list);
				break;
			case -2:
				notice(s_ChanServ, source,
					"%s ERROR: Channel %s is at it's maximum capicty for access entries.",
						list, chan);
				break;
			case -1:
				notice(s_ChanServ, source,
					"%s ERROR: %s \2%s\2 is already on the %s list for %s.",
						list, mask ? "Mask" : "Nick", nick, list, chan);
				break;
			case 1:
				notice(s_ChanServ, source,
					"%s \2%s\2 added to the %s list on channel %s.",
						mask ? "Mask" : "Nick", nick, list, chan);
				break;
			case 2:
				notice(s_ChanServ, source,
					"%s \2%s\2 changed to an %s on channel %s.",
						mask ? "Mask" : "Nick", nick, list, chan);
				break;
			default:
				notice(s_ChanServ, source,
					"%s ERROR: Unknown error, please try again later.", list);
				break;
		}
	} else if (sstrcmp(cmd, "DEL") == 0) {
		struct AccessList *a, *a_del=0;
		const char *nick = param;
		int dacc=0, i=0, z=0;

		if ((strspn(nick, "1234567890") == strlen(nick)) && ((dacc=atoi(nick)) > 0)) {
			for (z=0,a=ci->entries; a; z++, a=a->next) {
				if (a->level != level)
					continue;
				i++;
				if (dacc == i) {
					a_del = a;
					break;
				}
			}
			if (!a_del) {
				notice(s_ChanServ, source,
					"No such entry (#%d) on %s %s list.", dacc, chan, list);
				return;
			}

		} else {    /* not selecting numberized entry, using actual nick/mask */

			for (a = ci->entries; a; a = a->next) {
				if ((a->level == level) && sstrcmp(a->entry, nick) == 0) {
					a_del = a;
					break;
				}
			}
			if (!a_del) {
				notice(s_ChanServ, source,
					"No such entry '%s' on %s %s list.", nick, chan, list);
				return;
			}

		}
		notice(s_ChanServ, source,
			"Deleted '%s' from the %s list of channel %s", a_del->entry, list, chan);

		update_caccess(ci, 0, a_del->entry, ntrace, NULL, 0, acc, 0);

	} else if (sstrcmp(cmd, "WIPE") == 0) {
		struct AccessList *a=ci->entries, *a2;
		int entries=0;


		if (acc < 13) {   /* only founders */
			notice(s_ChanServ, source,
			"WIPE ERROR: You must have atleast Co-Founder access to use this command.");
			return;
		}

		while (a) {
			a2 = a->next;
			if (a->level == level) {
				if (a->prev)
					a->prev->next = a->next;
				else
					ci->entries = a->next;
				if (a->next)
					a->next->prev = a->prev;
				free(a->entry); free(a->instater); free(a);
				entries++;
			}
			a = a2;
		}

		send_sql("DELETE FROM caccess WHERE chan=%d AND level=%d", ci->id, level);

		if (entries)
			notice(s_ChanServ, source,
				"Removed all entries from the %s list of %s. (%d entries)", list, chan, entries);
		else
			notice(s_ChanServ, source,
				"%s ERROR: No %s entries found to wipe for channel %s.", list, list, chan);

	}

}

/************************************************************************************/

static void do_akick(const char *source)
{
	ChanInfo *ci;
	const char *chan = gettok(NULL, " ", CHANMAX);
	const char *cmd = gettok(NULL, " ", CMDMAX);
	const char *param = strtok(NULL, " ");
	const char *reason = strtok(NULL, "");
	int acc=0, isoper=0;
	char *ntrace;

	if ((!chan || !cmd) ||
		((!sstrcmp(cmd, "ADD") || !sstrcmp(cmd, "DEL")) && !param)) {
		notice(s_ChanServ, source,
			"Syntax: /%s AKICK <#channel> [ADD|DEL|LIST|WIPE] [<nick>] [reason]", s_ChanServ);
		notice(s_ChanServ, source,
		"\2/%s HELP AKICK\2 for more information.", s_ChanServ);

	} else if (!(ci=findchaninfo(chan))) {
		notice(s_ChanServ, source,
			"AKICK ERROR: Channel %s is not registered!", chan);



	} else if ((acc=get_access(finduser(source), chan, &ntrace, 0)) < 5 && !(isoper=is_oper(source))) {
		notice(s_ChanServ, source,
			"AKICK ERROR: You do not have access for this command on channel %s.", chan);

	} else if (isoper && sstrcmp(cmd, "LIST") != 0 && acc < 5) {
		notice(s_ChanServ, source,
			"AKICK ERROR: You do not have access for this command on channel %s.", chan);


	} else if (sstrcmp(cmd, "LIST") == 0) {
		struct AccessList *a;
		int i=0;
		char *lmask;

		for (a = ci->entries; a; a = a->next) {
			if (a->level < 0) {
				lmask=0;
				if (!i)
					notice(s_ChanServ, source,
						"AKICK list for \2%s\2:", chan);

				if (a->level == -2) {  /* nick entry */
					if (sql_select("name,usermask FROM nick WHERE id = %d", a->nickid)) {
						row = fetch(res);
						if (MASK_HOSTNAME)
							lmask = mask_host(row[1]);

						notice(s_ChanServ, source,
							"  %d) %s    (%s)  [Added by: %s]   %s%s%s",
							++i, row[0], (MASK_HOSTNAME && lmask && !isoper) ? lmask : row[1],
							a->instater, a->reason ? "(" : "", 
							a->reason ? a->reason : "", a->reason ? ")" : ""); 

					}
				} else {  /* mask entry */

					if (MASK_HOSTNAME)
						lmask = mask_host(a->entry);

					notice(s_ChanServ, source,
						"  %d) %s  [Added by: %s]  %s%s%s", ++i, 
						(MASK_HOSTNAME && lmask && !isoper) ? lmask : a->entry, a->instater,
					a->reason ? "(" : "", a->reason ? a->reason : "", a->reason ? ")" : ""); 
					if (lmask)
						free(lmask);

				}
			}
		}
		if (!i)
			notice(s_ChanServ, source, "No AKICK entries on record for channel %s", chan);
		else
			notice(s_ChanServ, source, "\2End of AKICK List\2 (%d entries found)", i);


	} else if (acc < 10) {
		notice(s_ChanServ, source,
			"AKICK ERROR: You do not have access for this command on channel %s.", chan);


/* User is only identified to channel (SET permissions only) */
	} else if (acc==15 && get_access(finduser(source), chan, &ntrace, 1) < 10) {
		notice(s_ChanServ, source,
			"AKICK ERROR: You may not manipulate the AKICK list because you are not identified "
			"to any nick on the access list of %s. Identifying to a channel only gives you "
			"access to the SET commands.", chan); 




	} else if (sstrcmp(cmd, "ADD") == 0) {
		User *u;
		char *mask, *maskp;
		int result=0, used_nick=0, regnick=0;


		if (reason && (strlen(reason) > 100)) {
			notice(s_ChanServ, source,
				"AKICK ERROR: Length of reason too long. Please try again with an "
				"abbreviated akick reason. (Chars: %d Limit: 100)", strlen(reason));
			return;
		}


		if (!(u = finduser(param))) {
			char *ptr = strchr(param, '!');
			if (ptr)
				*ptr++;

			if (!strchr(param, '!') && !strchr(param, '@')) {
				if (!(regnick=sql_select("usermask,name,id FROM nick WHERE name = '%s'", param))) {
					notice(s_ChanServ, source,
						"AKICK ERROR: Nick '%s' is not online or registered!", param);
					return;
				}
				row = fetch(res);
				mask = smalloc(strlen(row[0]) + 3);
				sprintf(mask, "*!%s", row[0]);
				used_nick=1;

			} else {

				if (strlen(param) > 60) {
					notice(s_ChanServ, source,
						"AKICK ERROR: Length of mask is too long!");
					return;

				} else if (!strchr(param, '!') || !strchr(param, '@')) {
					notice(s_ChanServ, source,
						"AKICK ERROR: Mask entries must be in the form of nick!user@host.");
					return;


				} else if (*param == '!' || *ptr == '@' || !check_valid_mask(ptr)) {
					notice(s_ChanServ, source,
					"AKICK ERROR: Invalid mask - Must be in the format of nick!user@host. "
					"(wildcards '*' are valid)");
					return;

				} else
					mask = sstrdup(param);
			}
		} else {
			char *tmask;

			if ((regnick=sql_select("usermask,name,id FROM nick WHERE name = '%s'", param))) {
				row = fetch(res);
				mask = sstrdup(row[0]);
			} else {
				tmask = create_mask(u);
				mask = smalloc(strlen(tmask) + 3);
				sprintf(mask, "*!%s", tmask);
				free(tmask);
			}
			used_nick=1;
		}

		if (used_nick && MASK_HOSTNAME)
			maskp = mask_host(mask);
		else
			maskp = mask;

		if (regnick)
			result = update_caccess(ci, -2, row[1], ntrace, reason, 0, acc, strtoul(row[2], NULL, 10));
		else
			result = update_caccess(ci, -1, mask, ntrace, reason, 0, acc, 0);


		switch(result)
		{
			case -5:
				notice(s_ChanServ, source,
					"AKICK ERROR: You may not manipulate the access of %s on channel %s.",
						param, chan);
				break;
			case -3:
				notice(s_ChanServ, source,
					"AKICK ERROR: Channel %s is at it's maximum capicty for AKICK entries.",chan);
				break;
			case -2:
				notice(s_ChanServ, source,
					"AKICK ERROR: Channel %s is at it's maximum capicty for access entries.",chan);
				break;
			case -1:
				if (regnick)
					notice(s_ChanServ, source,
					"AKICK ERROR: Nick \2%s\2 (%s) is already on the AKICK list for %s.",
					param, maskp ? maskp : mask, chan);

				else
					notice(s_ChanServ, source,
					"AKICK ERROR: Mask \2%s\2 %s%s%sis already on the AKICK list for %s.",
						maskp ? maskp : mask, used_nick ? "(" : "", used_nick ? param : "",
						used_nick ? ") " : "", chan);

				break;
			case 1:
				if (regnick) {
					notice(s_ChanServ, source,
						"Nick \2%s\2 (%s) added to the AKICK list on channel %s.",
							param, maskp ? maskp : mask, chan);
				} else
					notice(s_ChanServ, source,
						"Mask \2%s\2 added to the AKICK list on channel %s.",
							maskp ? maskp : mask, chan);
				mass_check_kick(chan);
				break;

			case 2:
				if (regnick) {
					notice(s_ChanServ, source,
						"Nick \2%s\2 (%s) changed to the AKICK list on channel %s.",
							param, maskp ? maskp : mask, chan);
					break;
				}

			default:
				notice(s_ChanServ, source,
					"AKICK ERROR: Unknown error, please try again later.");
				break;
		}

		free(mask);


	} else if (sstrcmp(cmd, "DEL") == 0) {
		struct AccessList *a, *a_del=NULL;
		const char *nick = param;
		char *mask=0;
		int accd=0, i=0, z=0;

		if ((strspn(nick, "1234567890") == strlen(nick)) && ((accd=atoi(nick)) > 0)) {
			for (z=0,a=ci->entries; a; z++, a=a->next) {
				if (a->level > -1)
					continue;
				i++;
				if (accd == i) {
					a_del = a;
					break;
				}
			}
			if (!a_del) {
				notice(s_ChanServ, source,
					"No such entry (#%d) on %s AKICK list.", accd, chan);
				return;
			}

		} else {    /* not selecting numberized entry, using actual nick/mask */

			for (a = ci->entries; a; a = a->next) {
				if ((a->level < 0) && sstrcmp(a->entry, nick) == 0) {
					a_del = a;
					break;
				}
			}
			if (!a_del) {
				notice(s_ChanServ, source,
					"No such entry '%s' on %s AKICK list.", nick, chan);
				return;
			}

		}
		if (MASK_HOSTNAME)
			mask = mask_host(a_del->entry);
		notice(s_ChanServ, source,
			"Deleted '%s' from the AKICK list of channel %s", 
			(MASK_HOSTNAME && mask) ? mask : a_del->entry, chan);
		if (mask)
			free(mask);

		update_caccess(ci, 0, a_del->entry, ntrace, NULL, 0, acc, 0);

	} else if (sstrcmp(cmd, "WIPE") == 0) {
		struct AccessList *a=ci->entries, *a2;
		int entries=0;


		if (acc < 13) {   /* only founders */
			notice(s_ChanServ, source,
			"WIPE ERROR: You must have atleast Co-Founder access to use this command.");
			return;
		}

		while (a) {
			a2 = a->next;
			if (a->level < 0) {
				if (a->prev)
					a->prev->next = a->next;
				else
					ci->entries = a->next;
				if (a->next)
					a->next->prev = a->prev;
				free(a->entry); free(a->instater); free(a);
				entries++;
			}
			a = a2;
		}

		send_sql("DELETE FROM caccess WHERE chan=%d AND level < 0", ci->id);

		if (entries)
			notice(s_ChanServ, source,
				"Removed all entries from the AKICK list of %s. (%d entries)", chan, entries);
		else
			notice(s_ChanServ, source,
				"AKICK ERROR: No AKICK entries found to wipe for channel %s.", chan);

	} else {
		notice(s_ChanServ, source,
			"Syntax: /%s AKICK <#channel> [ADD|DEL|LIST|WIPE] <nick>", s_ChanServ);
		notice(s_ChanServ, source,
			"\2/%s HELP AKICK\2 for more information.", s_ChanServ);
	}
}

/************************************************************************************/

static void do_clear(const char *source, int expel)
{
	Channel *c;
	ChanInfo *ci;
	User *u = finduser(source);
	const char *chan = gettok(NULL, " ", CHANMAX);
	const char *cmd = gettok(NULL, " ", CMDMAX);
	char *r_cmd="CLEAR";
	int acc=0;

	if (expel)
		r_cmd = "EXPEL";


	if (!u)
		return;

	else if (!chan || !cmd) {
		notice(s_ChanServ, source,
			"Syntax: /%s %s <#channel> <what> [<reason>]", s_ChanServ, r_cmd);
		notice(s_ChanServ, source,
			"\2/%s HELP %s\2 for more information.", s_ChanServ, r_cmd);

	} else if (!(ci=findchaninfo(chan)) && !expel) {
		notice(s_ChanServ, source,
			"CLEAR ERROR: Channel %s is not registered!", chan);

	} else if (!(c = findchan(chan))) {
		notice(s_ChanServ, source,
			"%s ERROR: Channel %s is not currently in use! (empty)", r_cmd, chan);


	} else if (!expel && is_frozen(chan)) {
		notice(s_ChanServ, source,
			"CLEAR ERROR: Channel %s is currently frozen!", chan);


	} else if (!expel && (acc=get_access(u, chan, NULL, 0)) < 5) {
		notice(s_ChanServ, source,
			"CLEAR ERROR: You do not have access for this command on channel %s.", chan);

	} else if (sstrcmp(cmd, "BANS") == 0) {
		struct ChanBans *cb = c->bans, *cb2;
		char *buf[4];
		int bans=0, banned = cb ? 1 : 0;

		while (cb) {
			cb2 = cb->next;
			buf[bans++] = sstrdup(cb->mask);
			if (bans==4) {
				send_cmd(s_ChanServ, "MODE %s -bbbb %s %s %s %s",
					chan, buf[0], buf[1], buf[2], buf[3]);
				free(buf[0]); free(buf[1]); free(buf[2]); free(buf[3]);
				bans=0;
			}
			free(cb->mask); free(cb);
			cb = cb2;
		}
		if (bans) {
			send_cmd(s_ChanServ, "MODE %s -b%s%s %s %s %s", chan,
				bans>1 ? "b" : "", bans>2 ? "b" : "", buf[0],
				bans>1 ? buf[1] : "", bans>2 ? buf[2] : "");
			free(buf[0]);
			if (bans>1)
				free(buf[1]);
			if (bans>2)
				free(buf[2]);
		}
		c->bans = NULL;
		c->activebans = 0;


		if (banned) {
			notice(s_ChanServ, source,
				"Cleared all bans on channel %s.", chan);
			send_cmd(s_ChanServ,
				"NOTICE @+%s :%s used %s CLEAR BANS for channel %s.",
				chan, source, s_ChanServ, chan);
			if (expel)
				globops(s_ChanServ, "%s used EXPEL BANS on channel %s.",
					opertrace(u,0), chan);

		} else
			notice(s_ChanServ, source,
				"There are currently no bans to clear on channel %s.", chan);

	} else if ((sstrcmp(cmd, "OPS") == 0) || (sstrcmp(cmd, "VOICES") == 0)) {
		struct ChanUsers *cu=c->ops, *cu2;
		char buf[(NICKMAX*6) + 6];
		int op=1, times=0, deop=0;

		if (sstrcmp(cmd, "VOICES") == 0) {
			op=0;
			cu=c->voices;
			deop = cu ? 1 : 0;
		}

		*buf=0;
		while (cu) {
			cu2 = cu->next;
			if ((!expel && get_access(cu->user, chan, NULL, 0) >= acc) ||
				(expel && cu->user == u)) {
				cu = cu2;
				continue;
			}

			deop=1;
			times++;
			if (*buf)
				strcat(buf, " ");
			strcat(buf, cu->user->nick);
			if (times==6) {
				send_cmd(s_ChanServ, "MODE %s -%s %s",
					chan, op ? "oooooo" : "vvvvvv", buf);
				*buf=0;
				times=0;
			}
			if (cu->prev)
				cu->prev->next = cu->next;
			else {
				if (op)
					c->ops = cu->next;
				else
					c->voices = cu->next;
			}
			if (cu->next)
				cu->next->prev = cu->prev;
			free(cu);
			cu = cu2;
		}


		if (times)
			send_cmd(s_ChanServ, "MODE %s -%s %s",
				chan, op ? "oooooo" : "vvvvvv", buf);


		if (deop) {
			notice(s_ChanServ, source,
				"All %s have been set -%s in channel %s.",
				op ? "ops that you out-rank" : "voices", op ? "o" : "v", chan);
			send_cmd(s_ChanServ,
				"NOTICE @%s :%s used %s CLEAR %s for channel %s.",
				chan, source, s_ChanServ, op ? "OPS" : "VOICES", chan);
			if (expel)
				globops(s_ChanServ, "%s used EXPEL %s on channel %s.",
					opertrace(u,0), op ? "OPS" : "VOICES", chan);

		} else {
			if (!expel)
				notice(s_ChanServ, source,
					"There are currently no %s to clear on channel %s.", 
					op ? "ops that you out-rank" : "voices", chan);
			else
				notice(s_ChanServ, source,
					"There are currently no ops to clear on channel %s.", chan);
		}

	} else if (sstrcmp(cmd, "MODES") == 0) {
		char buf[15];
		long modes=0;
		int key=0;

		*buf=0;

		if (ci) {
			if (sql_select("modeson FROM chan WHERE id = %d", ci->id)) {
				row = fetch(res);
				modes = atol(row[0]);
			}
		} else
			modes = 0;


		if ((!(modes & CMODE_C)) && (c->modes & CMODE_C)) {
			c->modes &= ~CMODE_C;
			strcat(buf, "c");
		}
		if ((!(modes & CMODE_I)) && (c->modes & CMODE_I)) {
			c->modes &= ~CMODE_I;
			strcat(buf, "i");
		}
		if ((!(modes & CMODE_K)) && (c->modes & CMODE_K)) {
			c->modes &= ~CMODE_K;
			strcat(buf, "k");
			key=1;
		}
		if ((!(modes & CMODE_L)) && (c->modes & CMODE_L)) {
			c->modes &= ~CMODE_L;
			strcat(buf, "l");
			c->chanlimit = 0;
		}
		if ((!(modes & CMODE_M)) && (c->modes & CMODE_M)) {
			c->modes &= ~CMODE_M;
			strcat(buf, "m");
		}
		if ((!(modes & CMODE_N)) && (c->modes & CMODE_N)) {
			c->modes &= ~CMODE_N;
			strcat(buf, "n");
		}
		if ((!(modes & CMODE_P)) && (c->modes & CMODE_P)) {
			c->modes &= ~CMODE_P;
			strcat(buf, "p");
		}
		if ((!(modes & CMODE_r)) && (c->modes & CMODE_r)) {
			c->modes &= ~CMODE_r;
			strcat(buf, "R");
		}
		if ((!(modes & CMODE_S)) && (c->modes & CMODE_S)) {
			c->modes &= ~CMODE_S;
			strcat(buf, "s");
		}
		if ((!(modes & CMODE_T)) && (c->modes & CMODE_T)) {
			c->modes &= ~CMODE_T;
			strcat(buf, "t");
		}
		if ((!(modes & CMODE_m)) && (c->modes & CMODE_m)) {
			c->modes &= ~CMODE_m;
			strcat(buf, "M");
		}
		if ((!(modes & CMODE_o)) && (c->modes & CMODE_o)) {
			c->modes &= ~CMODE_o;
			strcat(buf, "O");
		}

		if (!*buf)
			notice(s_ChanServ, source,
				"There are no applicable modes to CLEAR from channel %s.", chan);
		else {
			send_cmd(s_ChanServ, "MODE %s -%s %s", chan, buf,
				key ? c->chankey : "");
				
			if (key) {
				free(c->chankey);
				c->chankey = NULL;
			}

			notice(s_ChanServ, source,
				"Channel %s has been set -%s", chan, buf);
			if (expel)
				globops(s_ChanServ, "%s used EXPEL MODES on channel %s.",
					opertrace(u,0), chan);

		}



	} else if (!expel && acc < 10) {
			notice(s_ChanServ, source,
				"You currently do not have access to use CLEAR USERS for %s.", chan);

	} else if (sstrcmp(cmd, "USERS") == 0) {
		struct ChanUsers *cu=c->users, *cu2;
		const char *reason = strtok(NULL, "");
		Timeout *suspend;

		if (reason && strlen(reason) > 300) {
			notice(s_ChanServ, source,
				"EXPEL ERROR: Length of reason too long.");
			return;
		}

		suspend = add_timeout(CHAN_INHABIT, timeout_clear, 0);
		suspend->data = sstrdup(chan);


		if (ci)
			ci->flags |= CS_SUSPEND;
		send_cmd(s_ChanServ, "MODE %s +lb 1 *!*@*", chan);
		send_cmd(s_ChanServ, "JOIN %s", chan);

		while (cu) {
			cu2 = cu->next;
			if (expel && cu->user == u) {
				cu = cu->next;
				continue;
			}
			send_cmd(s_ChanServ, "KICK %s %s :%s%s%s", chan, cu->user->nick,
				reason ? "Mass kick: " : "",
				reason ? reason : "CLEAR USERS command used by ",
				reason ? "" : source);

			chan_deluser(cu->user, c);

			cu = cu2;
		}
		if (expel)
			globops(s_ChanServ, "%s used EXPEL USERS on channel %s.",
				opertrace(u,0), chan);

		else
			blog("CS_M", "[CS] C %s (%s!%s@%s) [USERS]",
				chan, source, u->username, u->host);
	} else {
		notice(s_ChanServ, source,
			"Invalid command '%s'. "
			"\2/%s HELP %s\2 for more information.", cmd, s_ChanServ, r_cmd);
	}
}


/************************************************************************************/

static void do_close(const char *source)
{
	CloseChans *cc;
	User *u = finduser(source);
	const char *chan = gettok(NULL, " ", CHANMAX);
	const char *expiry = strtok(NULL, " ");
	const char *reason = strtok(NULL, "");

	if (!u)
		return;

	if (!chan || ((*chan != '-' && *chan != '*') && (!expiry || !reason))) {
		notice(s_ChanServ, source,
			"Syntax: /%s CLOSE <#channel> <time> <reason>", s_ChanServ);
		notice(s_ChanServ, source,
			"\2/%s HELP CLOSE\2 for more information.",	s_ChanServ);

	} else if (sstrcmp(chan, "*") == 0) {
		int i, found=0;
		char *s=0;

		int num_rows=sql_select("cflags.name,reason.instater,reason.reason,cflags.closed,cflags.expiry "
			"FROM cflags,reason WHERE cflags.closed > 0 AND cflags.name=reason.name "
			"AND reason.type = 'CLOSED'");

		notice(s_ChanServ, source, "\2CLOSED chan listing:\2");


		for (i=0; i < num_rows; i++) {
			row = fetch(res);

			if (atol(row[4]))
				s = sstrdup(calcltime(atol(row[4]) - time(NULL)));

			notice(s_ChanServ, source,
				"%d) %s Placed: %s   Expires: %s",
					i+1, row[0],timebuf(atol(row[3])), 
					atol(row[4]) ? s : "Permanent.");
			notice(s_ChanServ, source,
				"   Reason: \"%s\" (%s)", row[2], row[1]);

			if (atol(row[4]))
				free(s);
			found++;
		}
		if (!num_rows)
			notice(s_ChanServ, source, "No channels are currently closed.");
		else
			notice(s_ChanServ, source, "\2End of CLOSED list. (%d chans found)", found);


	} else if (*chan == '-') {
		int found=0;
		*chan++;

		for (cc = closelist; cc; cc = cc->next) {
			if (sstrcmp(cc->chan, chan) == 0) {
				send_sql("UPDATE cflags SET closed=0 WHERE name='%s'", chan);
				send_sql("DELETE FROM reason WHERE name='%s' AND type='CLOSED'", chan);
				if (cc->prev)
					cc->prev->next = cc->next;
				else
					closelist = cc->next;
				if (cc->next)
					cc->next->prev = cc->prev;
				free(cc->chan); free(cc);
				found=1;
				break;
			}
		}
		if (!found)
			notice(s_ChanServ, source,
				"CLOSE ERROR: Channel %s is not closed!", chan);
		else {
			notice(s_ChanServ, source,
				"Channel %s is now re-opened.", chan);
			globops(s_ChanServ, "%s has re-opened the channel %s.", 
				opertrace(u,0), chan);
		}


	} else if (*chan != '#' && *chan != '&') {
		notice(s_ChanServ, source,
			"CLOSE ERROR: Invalid channel '%s'.", chan);


	} else if (strlen(reason) > 250) {
		notice(s_ChanServ, source,
			"CLOSE ERROR: Length of reason too long!");

	} else if ((strlen(expiry) > 8) || (atoi(expiry) < 0)) {
		notice(s_ChanServ, source,
			"CLOSE ERROR: Bogus expiry time. Try a more suitable expiry.");


	} else {
		int amount=0, second=0;
		char expbuf[64], *c=0, *expp;
		time_t exp = 0;

		for (cc = closelist; cc; cc = cc->next) {
			if (sstrcmp(cc->chan, chan) == 0) {
				notice(s_ChanServ, source,
					"Channel %s is already closed!", cc->chan);
				return;
			}
		}

		if (atol(expiry) && !strchr(expiry, 'd') && !strchr(expiry, 'h') && !strchr(expiry, 'm')) {
			second=1;
			if (atol(expiry) < 60)
				sprintf(expbuf, "%lds", atol(expiry));
			else if (atol(expiry) < 3600)
				sprintf(expbuf, "%ldm", atol(expiry)/60);
			else if (atol(expiry) < 86400)
				sprintf(expbuf, "%ldh", atol(expiry)/3600);
			else
				sprintf(expbuf, "%ldd", atol(expiry)/86400);
		}

		expp = sstrdup(expiry);
		amount = strtol(expiry, (char**)&expiry, 10);

		if (amount) {
			switch(*expiry) {
				case 'd' : amount *= 60*60*24; c="d"; break;
				case 'h' : amount *= 60*60; c="h"; break;
				case 'm' : amount *= 60; c="m"; break;
			}
			exp = amount + time(NULL);
		}



		if (sql_select("* FROM cflags WHERE name = '%s'", chan))
			send_sql("UPDATE cflags SET closed=%d WHERE name='%s'",
				time(NULL), chan);
		else
			send_sql("INSERT INTO cflags VALUES "
				"('%s', 0, 0, %d, 0, %d)", chan, time(NULL), exp);


		send_sql("INSERT INTO reason VALUES "
			"('%s','CLOSED','%s','%s')", 
			chan, reason, opertrace(u,1));

		cc = scalloc(sizeof(CloseChans), 1);
		cc->chan = sstrdup(chan);
		cc->expiry = exp;
		cc->next = closelist;
		if (closelist)
			closelist->prev = cc;
		closelist = cc;

		notice(s_ChanServ, source,
			"Channel %s is now %s CLOSED.", chan, exp ? "temporarily" : "Permanently");
		globops(s_ChanServ, "%s has %s CLOSED the channel %s.%s%s%s%s Reason: %s",
			opertrace(u,0), exp ? "Temporarily" : "Permanently", chan,
			exp ? " [Expiry: " : "", second ? expbuf : exp ? strtok(expp, c) : "",
			exp && !second ? c : "", exp ? "]" : "", reason);

		free(expp);
		mass_check_kick(chan);

	}
}

/************************************************************************************/

static void do_count(const char *source)
{
	const char *chan = gettok(NULL, " ", CHANMAX);
	User *u = finduser(source);
	ChanInfo *ci;

	if (!u)
		return;

	else if (!chan) {
		notice(s_ChanServ, source,
			"Syntax: /%s COUNT <#channel>", s_ChanServ);
		notice(s_ChanServ, source,
			"\2/%s HELP COUNT\2 for more information.",
			s_ChanServ);
		
	} else if (!(ci = findchaninfo(chan))) {
		notice(s_ChanServ, source,
			"COUNT ERROR: Channel %s is not registered!", chan);

	} else if ((get_access(u, chan, NULL, 0) < 3) && !is_oper(source)) {
		notice(s_ChanServ, source,
			"COUNT ERROR: You do not have access to this command for channel %s.", chan);
	} else {
#ifdef HAS_HALFOP
		int h=0;
#endif
		int f=0, s=0, aa=0, v=0, k=0;
		struct AccessList *a;

		for (a = ci->entries; a; a = a->next) {
			if (a->level == 3)
				v++;
#ifdef HAS_HALFOP
			else if (a->level == 4)
				h++;
#endif
			else if (a->level == 5)
				aa++;
			else if (a->level == 10)
				s++;
			else if (a->level == 13)
				f++;
			else if (a->level < 0)
				k++;
		}
		
#ifdef BAHAMUT
		notice(s_ChanServ, source,
			"\2%s Count:\2 AVOICE: %d  AOP: %d  SOP: %d  CF: %d  AKICK: %d",
			chan,v,aa,s,f,k);
#else
		notice(s_ChanServ, source,
			"\2%s Count:\2 AVOICE: %d  HOP: %d  AOP: %d  SOP: %d  CF: %d  AKICK: %d",
			chan,v,h,aa,s,f,k);
#endif
	}

}

/************************************************************************************/

static void do_delete(const char *source)
{
	ChanInfo *ci;
	const char *chan = gettok(NULL, " ", CHANMAX);

	if (!chan) {
		notice(s_ChanServ, source,
			"Syntax: /%s DELETE <#channel>", s_ChanServ);
		notice(s_ChanServ, source,
			"\2/%s HELP DELETE\2 for more information.",
			s_ChanServ);

	} else if (!(ci = findchaninfo(chan))) {
		notice(s_ChanServ, source,
			"DELETE ERROR: Channel %s is not registered!", chan);

	} else {
		User *u = finduser(source);

		del_ci_chan(chan);
		send_sql("DELETE FROM chan WHERE id = %d", ci->id);
		send_sql("DELETE FROM cflags WHERE name = '%s'", chan);
		send_sql("DELETE FROM reason WHERE name = '%s'", chan);
		send_sql("DELETE FROM caccess WHERE chan = %d", ci->id);

		notice(s_ChanServ, source,
			"Channel '%s' deleted from all database records.", chan);
		globops(s_ChanServ, "%s has deleted the channel '%s' from all database records.",
			opertrace(finduser(source),0), chan);

		blog("CS_O", "[CS] D+ %s (%s!%s@%s)", chan, source, u->username, u->host);

	}

}

/************************************************************************************/

static void do_deop(const char *source)
{
	ChanInfo *ci;
	const char *chan = gettok(NULL, " ", CHANMAX);
	const char *nick = gettok(NULL, " ", NICKMAX);

	if (!chan || !nick) {
		notice(s_ChanServ, source,
			"Syntax: /%s DEOP <#channel>", s_ChanServ);
		notice(s_ChanServ, source,
			"\2/%s HELP DEOP\2 for more information.",
			s_ChanServ);

	} else if (!findchan(chan)) {
		notice(s_ChanServ, source,
			"DEOP ERROR: Channel %s is not currently in use! (empty)", chan);

	} else if (!(ci=findchaninfo(chan))) {
		notice(s_ChanServ, source,
			"DEOP ERROR: Channel %s is not registered!", chan);

	} else if (is_frozen(chan)) {
		notice(s_ChanServ, source,
			"DEOP ERROR: Channel %s is frozen! %s will not comply.", chan, s_ChanServ);

	} else if (is_frozen(nick)) {
		notice(s_ChanServ, source,
			"DEOP ERROR: Nick '%s has been frozen by the network.", nick);


	} else if (get_access(finduser(source), chan, NULL, 0) < 5) {
		notice(s_ChanServ, source,
			"DEOP ERROR: You do not have access for this command on channel %s.", chan);

	} else if (!is_chanop(nick, chan)) {
		notice(s_ChanServ, source,
			"DEOP ERROR: Nick '%s' isn't a channel operator on %s!", nick, chan);

	} else if ((ci->flags & CS_PROTECT) && (sstrcmp(source, nick) != 0)) {
		notice(s_ChanServ, source,
			"DEOP ERROR: Channel %s has Op Protection enabled. You may not use DEOP on someone else's nick.", chan);

	} else {
		send_cmd(s_ChanServ, "MODE %s -o %s", chan, nick);
		send_cmd(s_ChanServ, "NOTICE @+%s :%s used %s DEOP: %s", chan, source, s_ChanServ, nick);
		chan_delop(chan, nick, 0);
	}

}

/************************************************************************************/

static void do_drop(const char *source)
{
	ChanInfo *ci;
	const char *chan = gettok(NULL, " ", CHANMAX);
	User *u = finduser(source);

	if (!chan) {
		notice(s_ChanServ, source,
			"Syntax: /%s DROP <#channel>", s_ChanServ);
		notice(s_ChanServ, source,
			"\2/%s HELP DROP\2 for more information.",
			s_ChanServ);

	} else if (!(ci=findchaninfo(chan))) {
		notice(s_ChanServ, source,
			"DROP ERROR: Channel %s is not registered!", chan);

	} else if (!is_founder(finduser(source), chan)) {
		notice(s_ChanServ, source,
			"DROP ERROR: Permission denied, you are not identified as the founder to %s", chan);

	} else if (is_frozen(chan)) {
		notice(s_ChanServ, source,
			"DROP ERROR: You may not drop %s because it is currently frozen.", chan);

	} else if (chan_closed(chan)) {
		notice(s_ChanServ, source,
			"DROP ERROR: You may not drop %s because it is currently closed.", chan);

	} else if (ci->flags & CS_HELD) {
		notice(s_ChanServ, source,
			"DROP ERROR: You may not drop %s because it is currently held.", chan);


	} else {
		blog("CS_O", "[CS] D %s (%s!%s@%s)", chan, source, u->username, u->host);
		notice(s_ChanServ, source,
			"Registration of channel %s has stopped. "
			"ChanServ will no longer monitor this channel and is now available "
			"to others to register.", chan);
		send_sql("DELETE FROM chan WHERE id = %d", ci->id);
		send_sql("DELETE FROM cflags WHERE name = '%s'", chan);
		send_sql("DELETE FROM reason WHERE name = '%s'", chan);
		send_sql("DELETE FROM caccess WHERE chan = %d", ci->id);
		del_ci_chan(chan);
	}
}

/************************************************************************************/

static void do_flags(const char *source, int flag)
{
	const char *chan = gettok(NULL, " ", CHANMAX);
	const char *reason = strtok(NULL, "");
	char *cmd, *lcmd;
	User *u = finduser(source);

	if (flag==FLAG_HOLD)       { cmd = "HELD"; lcmd = "held"; }
	else if (flag==FLAG_MARK)  { cmd = "MARKED"; lcmd = "marked"; }
	else if (flag==FLAG_FREEZE)  { cmd = "FROZEN"; lcmd = "frozen"; }
	else { lcmd = cmd = ""; }

	if (!chan || (!reason && ((*chan != '-') && sstrcmp(chan, "*")))) {
		notice(s_ChanServ, source,
			"Syntax: \2%s\2 [-]<\37#channel\37> <\37reason\37>",
                        flag==1 ? "HOLD" : flag==2 ? "MARK" : flag==3 ? "FREEZE" :"");
                notice(s_ChanServ, source,
                        "\2/%s HELP %s\2 for more information.", s_ChanServ,
                        flag==1 ? "HOLD" : flag==2 ? "MARK" : flag==3 ? "FREEZE" : "");

	} else if (*chan == '*')  {  /* list all chans */
		int num_rows=1, i, found=0;

		notice(s_ChanServ, source, "\2%s chan listing:\2", cmd);

		num_rows=sql_select("cflags.name,reason.instater,reason.reason,cflags.%s FROM cflags,reason "
			"WHERE cflags.%s > 0 AND cflags.name=reason.name AND reason.type = '%s'",
			lcmd,lcmd,cmd);

		for (i=0; i < num_rows; i++) {
			row = fetch(res);
			notice(s_ChanServ, source,
				"%d) %s  [Added by: %s] Reason: \"%s\"   Placed: %s",
					i+1, row[0],row[1],row[2],timebuf(atol(row[3])));
			found++;
		}
		if (!num_rows)
			notice(s_ChanServ, source, "No channels are currently %s.", cmd);
		else
			notice(s_ChanServ, source, "\2End of %s list. (%d chans found)", cmd, found);
		free(lcmd);

	} else if (*chan == '-') {
		*chan++;
		if (!sql_select("name FROM cflags WHERE name = '%s' AND %s > 0", chan, lcmd)) {
			notice(s_ChanServ, source,
				"%s ERROR: Channel '%s' is not %s!", chan, cmd);
		} else {
/* If all flags are now set to 0 expire_chans will clean this up on next update. */
			send_sql("UPDATE cflags SET %s = 0 WHERE name = '%s'", lcmd, chan);
			send_sql("DELETE from reason WHERE name = '%s' AND type = '%s'",
                                chan, cmd);
			if (flag==FLAG_HOLD) {
				ChanInfo *ci = findchaninfo(chan);

				if (ci && !(ci->flags & CS_HELD)) {
					ci->flags &= ~CS_HELD;
					send_sql("UPDATE chan SET flags = %d WHERE id = %d",
						ci->flags, ci->id);
				}

			} else if (flag==FLAG_FREEZE)
				freeze(chan, DEL);

			blog("CS_F", "[CS] %s- (%s!%s@%s)",
				flag==FLAG_HOLD ? "H" : flag==FLAG_MARK ? "M" : flag==FLAG_FREEZE ? "F" : "",
				u->os_nick, u->username, u->host);

			notice(s_ChanServ, source,
				"Channel '\2%s\2' is no longer %s.", chan, cmd);
			globops(s_ChanServ, "%s has removed the %s flag from channel '%s'",
				opertrace(u,0), cmd, chan);
		}

	} else if (strlen(reason) > 255) {
		notice(s_ChanServ, source,
			"ERROR: Your reason must not exceed 255 characters in length.");

	} else if (!findchaninfo(chan) && flag != FLAG_FREEZE) {
		notice(s_ChanServ, source,
			"ERROR: Channel '%s' is not registered!", chan);

	} else {
		if (sql_select("%s FROM cflags WHERE name='%s'", lcmd, chan)) {
			row = fetch(res);
				if (atol(row[0]) > 0) {
					notice(s_ChanServ, source,
						"ERROR: Channel '\2%s\2' is already %s!",
						chan, cmd);
					return;
				} else {
					send_sql("UPDATE cflags SET %s=%d WHERE name='%s'",
						lcmd,time(NULL),chan);
				}
		} else {
                                send_sql("INSERT INTO cflags VALUES "
                                        "('%s','%d','%d',0,'%d',0)", chan,
                                        flag==FLAG_HOLD ? time(NULL) : 0,
                                        flag==FLAG_MARK ? time(NULL) : 0,
                                        flag==FLAG_FREEZE ? time(NULL) : 0);
		}

		if (flag==FLAG_HOLD) {
				ChanInfo *ci = findchaninfo(chan);

				if (ci && !(ci->flags & CS_HELD)) {
	                                ci->flags |= CS_HELD;
	                                send_sql("UPDATE chan SET flags = %d WHERE id = %d", 
						ci->flags, ci->id);
				}
		} else if (flag==FLAG_FREEZE)
			freeze(chan, ADD);


		send_sql("INSERT INTO reason VALUES "
			"('%s','%s','%s','%s')", chan, cmd, reason,
			opertrace(finduser(source),1));
		blog("CS_F", "[CS] %s+ %s (%s!%s@%s)",
			flag==FLAG_HOLD ? "H" : flag==FLAG_MARK ? "M" : flag==FLAG_FREEZE ? "F" : "",
			chan, u->os_nick, u->username, u->host);
		notice(s_ChanServ, source,
			"Channel '\2%s\2' added to %s list", chan, cmd);
		globops(s_ChanServ, "%s has set the %s flag for channel '%s'.",
			opertrace(u,0), cmd, chan);

	}

}

/************************************************************************************/

static void do_flist(const char *source)
{
	char *cmd = strtok(NULL, " ");

	if (!cmd || (sstrcmp(cmd, "HELD") && sstrcmp(cmd, "MARKED") &&
		sstrcmp(cmd, "CLOSED") && sstrcmp(cmd, "FROZEN"))) {
		notice(s_ChanServ, source,
			"Syntax: \2FLIST\2 [HELD|MARKED|CLOSED|FROZEN]");
		notice(s_ChanServ, source,
			"\2/%s HELP FLIST\2 for more information.", s_ChanServ);

	} else if (sstrcmp(cmd, "CLOSED") == 0) { /* forward to do_close() for listing */
		char *parm = sstrdup("* *");
		strtok(parm, " ");
		do_close(source);
		free(parm);

	} else {
		int num_rows=0, i, found=0;

		notice(s_ChanServ, source, "\2%s chan listing:\2", strupper(cmd));

		num_rows=sql_select("cflags.name,reason.instater,reason.reason,cflags.%s FROM cflags,reason "
			"WHERE cflags.%s > 0 AND cflags.name=reason.name AND reason.type = '%s'",
			strlower(cmd),cmd,strupper(cmd));


		for (i=0; i < num_rows; i++) {
			row = fetch(res);
			notice(s_ChanServ, source,
				"%d) %s  [Added by: %s] Reason: \"%s\"   Placed: %s",
					i+1, row[0],row[1],row[2],timebuf(atol(row[3])));

			found++;
		}
		if (!num_rows)
			notice(s_ChanServ, source, "No channels are currently %s.", cmd);
		else
			notice(s_ChanServ, source, "\2End of %s list. (%d chans found)", cmd, found);
	}

}

/************************************************************************************/

static void do_getpass(const char *source)
{
	ChanInfo *ci;
	const char *chan = gettok(NULL, " ", CHANMAX);


#ifdef USE_ENCRYPTION
	notice(s_ChanServ, source,
		"GETPASS has been disabled by the network.");
	return;
#endif


	if (!ENABLE_GETPASS) {
		notice(s_ChanServ, source,
			"GETPASS has been disabled by the network.");

	} else if (!chan) {
		notice(s_ChanServ, source,
			"Syntax: /%s GETPASS <#channel>", s_ChanServ);
		notice(s_ChanServ, source,
			"\2/%s HELP GETPASS\2 for more information.",
			s_ChanServ);

	} else if (sql_select("instater FROM reason,cflags WHERE cflags.name='%s' AND "
		"cflags.name=reason.name AND MARKED > 0", chan)) {
		row = fetch(res);
		notice(s_ChanServ, source,
			"GETPASS ERROR: Channel %s has been marked by '%s' and may not be getpassed!",
			chan, row[0]);



	} else if (!(ci = findchaninfo(chan))) {
		notice(s_ChanServ, source,
			"GETPASS ERROR: Channel %s is not registered!", chan);

	} else if (!sql_select("pass,flags FROM chan WHERE id = %d", ci->id)) {
		notice(s_ChanServ, source,
			"GETPASS ERROR: Channel %s is not registered!", chan);

		if (debug)
			log("Channel %s (%d) found in memory but non-existent in database!", chan, ci->id);

	} else {
		User *u = finduser(source);
		row = fetch(res);

		notice(s_ChanServ, source,
			"Channel password retrieved for %s: \2%s\2", chan, row[0]);
		globops(s_ChanServ, "%s used GETPASS on channel %s.",
			opertrace(finduser(source),0), chan);

		blog("CS_S", "[CS] G %s (%s!%s@%s)", chan, source, u->username, u->host);

	}

}


/************************************************************************************/

static void do_register(const char *source)
{
	NickInfo *ni;
	const char *chan = gettok(NULL, " ", CHANMAX);
	const char *pass = strtok(NULL, " ");
	const char *desc = strtok(NULL, "");
	User *u;


	if (!chan || !pass || !desc) {
		notice(s_ChanServ, source,
			"Syntax: \2REGISTER\2 <channel> <password> <description>");
		notice(s_ChanServ, source,
			"\2/%s HELP REGISTER\2 for more information.", s_ChanServ);

	} else if (!(u = finduser(source))) {
		if (debug)
			log("%s registration failed! %s not online!", s_ChanServ, source);
		notice(s_ChanServ, source, "Sorry, registration failed.");

	} else if (*chan != '#') {
		if (*chan == '&')
			notice(s_ChanServ, source,
				"REGISTER error: Local channels may not be registered.");
		else
			notice(s_ChanServ, source,
				"REGISTER error: This channel is not valid for registration.");

	} else if (chan[1] == 0) {
		notice(s_ChanServ, source,
			"REGISTER error: This channel is not valid for registration.");

	} else if (sstrcmp(source, pass)==0 || sstrcmp(chan, pass)==0 ||
		strlen(pass) < 5) {
		notice(s_ChanServ, source,
			"REGISTER error: Please try again with a more obscure password. "
			"Passwords should be at least five characters in length, should not "
			"be something easily guessed (e.g. your real name or nick), and cannot "
			"contain the space or tab characters. "
			"\2/%s HELP REGISTER\2 for more information.", s_ChanServ);

	} else if (strlen(pass) > PASSMAX) {
		notice(s_ChanServ, source,
			"REGISTER error: Sorry, channel passwords may only be up to %d characters in length.",
			PASSMAX);

	} else if (check_escape_codes(pass)) {
		notice(s_ChanServ, source,
			"REGISTER error: Your password has invalid characters. Please try again with "
			"common characters and don't use special colour/attribute codes.");


	} else if (strlen(desc) > 200) {
		notice(s_ChanServ, source,
			"REGISTER error: Channel description exceedingly long. Please try a shorter one.");

	} else if (time(NULL) < (u->lastreg + NS_REG_DELAY)) {
		notice(s_ChanServ, source, NS_REGISTER_DELAY);

	} else if (!(ni = findnick(source))) {
		notice(s_ChanServ, source,
			"REGISTER error: You must be actively using a registered nick before "
			"attempting to founder a channel.");

	} else if (!is_identified(source, source)) {
		notice(s_ChanServ, source,
			"REGISTER error: You must first be identified to your active nick before "
			"attempting to founder a channel.");

	} else if (AUTH_EMAIL && NSECURE && !is_authed(source)) {
		notice(s_ChanServ, source,
			"REGISTER error: You must first authorize to your active nick before "
			"attempting to founder a channel.");

	} else if (!is_chanop(source, chan)) {
		notice(s_ChanServ, source,
			"REGISTER error: You must be a channel operator in '%s' to be able to register "
			"the channel.", chan);

	} else if (chan_closed(chan)) {
		notice(s_ChanServ, source,
			"REGISTER error: Channel '%s' is closed and may not be registered. ", chan);

	} else if (is_frozen(chan)) {
		notice(s_ChanServ, source,
			"REGISTER error: Channel '%s' is frozen and may not be registered. ", chan);


	} else if (findchaninfo(chan)) {
		notice(s_ChanServ, source,
			"REGISTER ERROR: Channel '%s' is already registered!", chan);


	} else if (MAX_CHANS && (check_maxchans(ni->id) >= MAX_CHANS) && !is_on_master_list(source)) {
		notice(s_ChanServ, source, MAX_CHANS_REACHED, source);

	} else {     /* register channel */
		ChanInfo *ci;
		long modeson=0, modesoff=0;
		long flags=0, id=0;
#ifndef USE_ENCRYPTION
		const char *passwd = pass;
#else
		char passwd[PASSMAX];
		encryptm(pass, strlen(pass), passwd, PASSMAX, 1);
#endif

		if (AUTH_EMAIL) {
			if (!sql_select("email FROM nick WHERE id = %d", ni->id)) {
				notice(s_ChanServ, source, "Sorry, registeration failed.");
				return;
			} else {
				row = fetch(res);
				if (row[0] == NULL) {
					notice(s_ChanServ, source,
						"You may not register a channel without a valid "
						"e-mail address attached to your nick.");
					if (debug)
						log("User %s tried to register %s without an e-mail!",
							source, chan);
					return;
				}
			}
		}

		modeson |= def_mlocks(1);
		modesoff |= def_mlocks(0);

		if (DEF_CS_KEEPTOPIC)
			flags |= CS_KEEPTOPIC;
		if (DEF_CS_IDENT)
			flags |= CS_IDENT;
#ifdef USE_ENCRYPTION
		flags |= CS_ENCRYPTED;
#endif

		ci = add_ci_chan(chan, source, 0);

		u->lastreg = time(NULL);

		ci->id = id = generate_primary_id("chan");
		ci->fid = ni->id;
		ci->flags = flags;

		send_sql("INSERT INTO chan VALUES "
			"(%d,'%s','%s','%s',%d,'%s',NULL,NULL,'%d','%d',NULL,NULL,%d,%d,%d,0,NULL,NULL,NULL,0,5)",
			id, chan, passwd, source, ni->id, desc, time(NULL), time(NULL), modeson,modesoff, flags);

		notice(s_ChanServ, source, CS_REGISTER, chan, source);
		notice(s_ChanServ, source, CS_REGISTER2, pass);
		if (strlen(DEF_MLOCKS) > 1)
			notice(s_ChanServ, source, CS_REGISTER3, chan, DEF_MLOCKS);

		blog("CS_R", "[CS] R %s (%s!%s@%s)",
			chan, source, u->username, u->host);
		add_id_chan(u, ci, 1);
		check_modes(findchan(chan), ci);
	}
}

/************************************************************************************/

static void help(const char *source)
{
	char *param = gettok(NULL, "", CMDMAX);
	char *s, *topic;
	int level=0;

	if (param) {
		s = sstrdup(param);
		topic = strtok(s, " ");
		if (topic) {
			int i=0;
			if ((i=check_cs_command(topic)) > -1)
				level = commands[i].access;
		}
		free(s);
	}
#ifdef HAS_HALFOP
	if (!param || sstrcmp(param, "HELP") == 0)
		do_help(s_ChanServ, source, "help2", level);
	else
		do_help(s_ChanServ, source, param, level);
#else
	do_help(s_ChanServ, source, param, level);
#endif
}

/************************************************************************************/

static void ohelp(const char *source) 
{

	Command *c;
	int shown=0;
	unsigned int level;
	User *u = finduser(source);

	if (!u)
		return;

	notice(s_ChanServ, source,
		"This is a listing of %s's privileged commands.", s_ChanServ);
	notice(s_ChanServ, source,
		"Your current access level is: %s.",
		u->os_level==0 ? "IRC Operator" :
		u->os_level==1 ? "Services Operator" :
		u->os_level==2 ? "Services Administrator" :
		u->os_level==3 ? "Services Root" :
		u->os_level>3 ? "Services Root Master" : "");


	notice(s_ChanServ, source, "For a detailed explanation for a command, type:");
	notice(s_ChanServ, source, "\2/%s HELP <command>", s_ChanServ);
	for (level=1; level < 5; level++, shown=0) {
		notice(s_ChanServ, source, " ");
		notice(s_ChanServ, source, "%s (%s)",
			level==1 ? "Commands available for \2IRC Operators\2:" :
			level==2 ? "Commands available for \2Services Operators\2:" :
			level==3 ? "Commands available for \2Services Administrators\2:" :
			level>3 ? "Commands available for \2Services Roots\2:" : "",
			u->os_level+1 >= level ? "Access Granted" : "Access Denied");
		notice(s_ChanServ, source, " ");

		for (c = commands; c->cmd; c++) {
			if (c->access == level) {
				notice(s_ChanServ, source, "    \2%s\2   %s\n", c->cmd, c->desc);
				shown++;
			}
		}
		if (!shown)
			notice(s_ChanServ, source, "    No commands available.");
	}
	notice(s_ChanServ, source, " ");
	notice(s_ChanServ, source, "Note: Most of these commands are logged!");
	notice(s_ChanServ, source, "*** END OF HELP ***");
}

/************************************************************************************/

static void do_identify(const char *source)
{
	ChanInfo *ci;
	const char *chan = gettok(NULL, " ", CHANMAX);
	const char *pass = strtok(NULL, " ");
	User *u = finduser(source);
	int reg=0;

	if (!u)
		return;

	if (!chan || !pass) {
		notice(s_ChanServ, source,
			"Syntax: /%s IDENTIFY <#channel> <password>", s_ChanServ);
		notice(s_ChanServ, source,
			"\2/%s HELP IDENTIFY\2 for more information.",
			s_ChanServ);
		return;
	}

	if (!(ci = findchaninfo(chan)))
		notice(s_ChanServ, source,
			"IDENTIFY ERROR: Channel '%s' is not registered!", chan);

	else if (!(reg = sql_select("pass FROM chan WHERE id = %d", ci->id))) {

		notice(s_ChanServ, source,
			"IDENTIFY ERROR: Channel '%s' is not registered!", chan);

		if (debug)
			log("Channel %s (%d) found in memory but non-existent in database!", chan, ci->id);



	} else if (reg) {
		row = fetch(res);
#ifdef USE_ENCRYPTION
		if (check_password(pass, row[0]) < 1) {
#else
		if (strcmp(row[0], pass) != 0) {  /* case sensitive */
#endif
			notice(s_ChanServ, source, CS_BAD_PASS, chan);
			blog("CS_I", "[CS] I* %s (%s!%s@%s)",
				chan, source, u->username, u->host);
			bad_password(u, source);
		} else {
			notice(s_ChanServ, source, CS_IDENTIFY, chan);
			add_idchan(u, findchaninfo(chan));
		}
	}
}

/************************************************************************************/

static char *show_coptions(char *buf, long flags, const char *chan)
{

	ChanInfo *ci = findchaninfo(chan);

	*buf=0;

	if (flags & CS_IDENT)
		strcat(buf, "Ident");

	if (flags & CS_KEEPTOPIC) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Topic Retention");
	}

	if (flags & CS_LEAVEOPS) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Leave Ops");
	}

	if (flags & CS_NOTIFY) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Notify");
	}

	if (flags & CS_OPGUARD) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Op Guard");
	}

	if (flags & CS_PRIVATE) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Private");
	}

	if (flags & CS_PROTECT) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Op Protect");
	}

	if (flags & CS_RESTRICTED) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Restricted");
	}

	if (flags & CS_TRIMBAN) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Trim Bans");
	}

	if (flags & CS_UNSECURE) {
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, "Unsecure");
	}


	if (ci && ci->topiclock > 0) {
		int tl = ci->topiclock;
		if (*buf)
			strcat(buf, ", ");
		strcat(buf, 
			tl==3  ? "Topic Lock(+v)" :
			tl==5  ? "Topic Lock(A)" :
			tl==10 ? "Topic Lock(S)" :
			tl==13 ? "Topic Lock(CF)" :
			tl==15 ? "Topic Lock(F)" : "");
	}
	if (!*buf)
		strcat(buf, "None");
	strcat(buf, ".");

	return buf;
}


/************************************************************************************/

#define INFO_NAME	1
#define INFO_PASS	2
#define INFO_FOUNDER	3
#define INFO_DESC	5
#define INFO_URL        6
#define INFO_WELCOME    7
#define INFO_TIME	8
#define INFO_LAST	9
#define INFO_KEY        10
#define INFO_FLAGS      14
#define INFO_TOPICLOCK  15
#define INFO_SUCCESSOR  16
#define INFO_MLEVEL     20

static void do_info(const char *source)
{
	ChanInfo *ci;
	const char *chan = gettok(NULL, " ", CHANMAX);
	long ltime;
	int mlevel=0;


	if (!chan) {
		notice(s_ChanServ, source,
			"Syntax: /%s INFO <#channel>", s_ChanServ);
		notice(s_ChanServ, source,
			"\2/%s HELP INFO\2 for more information.",
			s_ChanServ);

	} else if (chan_closed(chan)) {
		notice(s_ChanServ, source,
			"INFO: Channel '%s' is CLOSED and may not be used.", chan);

	} else if (!(ci = findchaninfo(chan)) ||
		!sql_select("* FROM chan,nick WHERE chan.id = %d AND nick.id=chan.fid", ci->id)) {
		notice(s_ChanServ, source,
			"INFO ERROR: Channel \2%s\2 is not registered!", chan);

	} else {
		char buf[512];
		long flags;
		int num_rows;
		char *mask=0;

		row = fetch(res);

		flags = atol(row[INFO_FLAGS]);
		mlevel = atoi(row[INFO_MLEVEL]);


		notice(s_ChanServ, source,
			"Information for channel \2%s\2:", row[INFO_NAME]);


		if (MASK_HOSTNAME)
			mask = mask_host(row[27]);
		notice(s_ChanServ, source,
			"         Founder: %s (%s)", row[INFO_FOUNDER],
			(MASK_HOSTNAME && !is_oper(source) && mask) ? mask : row[27]);
		if (mask)
			free(mask);


		notice(s_ChanServ, source,
			"       Successor: %s", row[INFO_SUCCESSOR] ? row[INFO_SUCCESSOR] : "<None>");

		notice(s_ChanServ, source,
			"     Description: %s", row[INFO_DESC]);


		notice(s_ChanServ, source,
			"      Registered: %s", timebuf(atol(row[INFO_TIME])));


		ltime = (time(NULL) - atol(row[INFO_LAST]));
		if (ltime > 86400)
			notice(s_ChanServ, source,
				"     Last Opping: %s [over %d day%s]", timebuf(atol(row[INFO_LAST])),
				ltime/86400, (ltime/86400 == 1) ? "" : "s");
		else if (ltime > 3600)
			notice(s_ChanServ, source,
				"     Last Opping: %s [over %d hour%s]", timebuf(atol(row[INFO_LAST])),
				ltime/3600, (ltime/3600 == 1) ? "" : "s");
		else
			notice(s_ChanServ, source,
				"     Last Opping: %s [under 1 hour]", timebuf(atol(row[INFO_LAST])));


		notice(s_ChanServ, source,
			"        Time Now: %s", timebuf(0));


		if (row[INFO_WELCOME])
			notice(s_ChanServ, source,
			"     Welcome Msg: %s", row[INFO_WELCOME]);

		if (row[INFO_URL])
			notice(s_ChanServ, source,
			"             URL: %s", row[INFO_URL]);

		if (row[INFO_KEY] && is_founder(finduser(source), chan))
			notice(s_ChanServ, source,
			"        Key lock: %s", row[INFO_KEY]);


		notice(s_ChanServ, source,
			"         Options: %s", show_coptions(buf, flags, chan));

		notice(s_ChanServ, source,
			"      Mode Locks: %s", get_cmodes(ci));


		notice(s_ChanServ, source,
			"      Memo Level: %s",
				!mlevel ? "Disabled" : mlevel==3 ? "Auto-Voice" : mlevel==5 ? "Auto-Op" :
				mlevel==10 ? "Super-Op" : mlevel==13 ? "Co-Founder" : "Founder");



		if (is_oper(source)) {
			if ((flags & CS_LEVELOPER || flags & CS_LEVELSOP || flags & CS_LEVELSA
				|| flags & CS_LEVELSRA || flags & CS_LEVELMASTER))
				notice(s_ChanServ, source,
					"  * Restricted to %s",
					flags & CS_LEVELOPER ? "oper access." :	
					flags & CS_LEVELSOP ? "Services Operator (SOP) access." :
					flags & CS_LEVELSA ? "Services Adminstrator (SA) access." :
					flags & CS_LEVELSRA ? "Services Root Access (SRA)." :
					flags & CS_LEVELMASTER ? "Services Root Master access." : "oper access.");



			if (is_oper(source) && (num_rows=sql_select(
				"type,instater FROM reason WHERE name = '%s'", chan))) {
				int i;
				for (i=0; i < num_rows; i++) {
					row = fetch(res);
					notice(s_ChanServ, source,
						"  * This channel has been flagged as %s by %s.",
						row[0], row[1]);
				}
			}
		} else {
			if (is_frozen(chan))
				notice(s_ChanServ, source,
					"  * This channel has been frozen by the network.");
		}


		if (flags & CS_OFFICIAL)
			notice(s_ChanServ, source,
				"  * This is an official channel of %s.", NETWORK_NAME);

		notice(s_ChanServ, source,
			"\2***  End of Info  ***\2");
	}
}

/************************************************************************************/

static void do_invite(const char *source)
{
	const char *chan = gettok(NULL, " ", CHANMAX);
	User *u = finduser(source);

	if (!chan) {
		notice(s_ChanServ, source,
			"Syntax: /%s INVITE <#channel>", s_ChanServ);
		notice(s_ChanServ, source,
			"\2/%s HELP INVITE\2 for more information.",
			s_ChanServ);


	} else if (!findchan(chan)) {
		notice(s_ChanServ, source,
			"INVITE ERROR: Channel '%s' is not currently in use!", chan);

	} else if (!findchaninfo(chan)) {
		notice(s_ChanServ, source,
			"INVITE ERROR: Channel '%s' is not registered!", chan);

	} else if (get_access(u, chan, NULL, 0) < 3) {
		notice(s_ChanServ, source,
			"INVITE ERROR: Access denied for channel %s.", chan);

	} else {
		struct UserChans *cl;
		for (cl = u->chans; cl; cl = cl->next) {
			if (sstrcmp(cl->chan->name, chan) == 0) {
				notice(s_ChanServ, source,
					"INVITE ERROR: You're already in channel %s!", chan);
				return;
			}
		}

		send_cmd(s_ChanServ, "INVITE %s %s", source, chan);
	}
}


/************************************************************************************/

static void do_level(const char *source)
{
	ChanInfo *ci;
	User *u = finduser(source);
	const char *chan = gettok(NULL, " ", CHANMAX);
	const char *level = strtok(NULL, " ");
	char *ntrace;

	if (!chan || !level) {
		notice(s_ChanServ, source,
			"Syntax: /%s LEVEL <#channel> <access>", s_ChanServ);
		notice(s_ChanServ, source,
			"\2/%s HELP LEVEL\2 for more information.",
			s_ChanServ);

	} else if (!(ci = findchaninfo(chan))) {
		notice(s_ChanServ, source,
			"LEVEL ERROR: Channel %s is not registered!", chan);

	} else if (get_access(u, chan, &ntrace, 0) < 13) {
		notice(s_ChanServ, source,
			"LEVEL ERROR: Permission denied. You must be atleast a co-founder of channel %s "
			"to perform this command.", chan);

	} else if (sstrcmp(level, "OPER") && sstrcmp(level, "SOP") && sstrcmp(level, "SA")
		&& sstrcmp(level, "SRA") && sstrcmp(level, "MASTER") && sstrcmp(level, "NONE")) {
		notice(s_ChanServ, source,
			"LEVEL ERROR: Invalid level '%s'. Only OPER, SOP, SA, SRA, and MASTER are valid.", level);

	} else {

		if (ci->flags & CS_LEVELOPER)
			ci->flags &= !CS_LEVELOPER;
		if (ci->flags & CS_LEVELSOP)
			ci->flags &= !CS_LEVELSOP;
		if (ci->flags & CS_LEVELSA)
			ci->flags &= !CS_LEVELSA;
		if (ci->flags & CS_LEVELSRA)
			ci->flags &= !CS_LEVELSRA;
		if (ci->flags & CS_LEVELMASTER)
			ci->flags &= !CS_LEVELMASTER;

		ci->flags |= 
			!sstrcmp(level, "OPER") ? CS_LEVELOPER :
			!sstrcmp(level, "SOP") ? CS_LEVELSOP :
			!sstrcmp(level, "SA") ? CS_LEVELSA :
			!sstrcmp(level, "SRA") ? CS_LEVELSRA :
			!sstrcmp(level, "MASTER") ? CS_LEVELMASTER : ci->flags;

		notice(s_ChanServ, source,
			"Level for channel %s set to \2%s\2", chan, level);
		send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);
	}

}

/************************************************************************************/

static void do_list(const char *source)
{
	char *param = gettok(NULL, "", 100);
	char *ptr;
	int num_rows=0, isoper = is_oper(source);


	if (((ALLOW_LIST==2 && !is_oper(source))) || !ALLOW_LIST) {
		notice(s_ChanServ, source,
			"Sorry, the network has disabled the LIST command.");
		return;
	}

	notice(s_ChanServ, source, "LIST of chans (Search paramater = %s)",
		param ? param : "*");


        if (param) {
		if (strlen(param) > CHANMAX) {
			notice(s_ChanServ, source, "ERROR: List paramater too long");
			return;
		}

		ptr = param;
		while (*ptr) {
		if (*ptr == '*')
			*ptr = '%';
			*ptr++;
		}

	} else
		param = "%";

	num_rows=sql_select("name,founder,descr,flags FROM chan "
		"WHERE name LIKE '%s' AND flags <> -1 ORDER BY name", param);

	if (num_rows) {
		unsigned private=0;
		int i, shown=0;
		for (i=0; i < num_rows && shown < LIST_MAX; i++) {
			row = fetch(res);
			if (atol(row[3]) & CS_PRIVATE) {
				if (!isoper) {	private++; continue; }
				private++;
			}
			shown++;
			notice(s_ChanServ, source, "%d) %-30s Founder: %s", shown, row[0], row[1]);
			notice(s_ChanServ, source, "   Desc: %s", row[2]);

		}
		notice(s_ChanServ, source,
			"Displayed %d of %d match results. (%d private) %s",
			shown, num_rows, private, LIST_MAX<num_rows ?
			"Exceeded maximum result limit, please try refining your search pattern" : "");
	} else {
		notice(s_ChanServ, source, "No registered channels for that match pattern.");
		return;
	}
	notice(s_ChanServ, source, "*** End of list ***");

}


/************************************************************************************/

static void do_official(const char *source)
{
	ChanInfo *ci;
	const char *chan = gettok(NULL, " ", CHANMAX);
	User *u = finduser(source);

	if (!u)
		return;

	else if (!chan) {
		notice(s_ChanServ, source,
			"Syntax: /%s OFFICIAL [-]<#channel>", s_ChanServ);
		notice(s_ChanServ, source,
			"\2/%s HELP OFFICIAL\2 for more information.",
			s_ChanServ);

	} else if (sstrcmp(chan, "*") == 0) {
		int i=1;
		notice(s_ChanServ, source,
			"\2OFFICIAL channel listing\2:");

		ci = official_list(NULL, 3);
		while (ci != NULL) {
			notice(s_ChanServ, source,
				" %d)  %s", i++, ci->name);
			ci = official_list(ci, 3);
		}
		notice(s_ChanServ, source, "\2End of OFFICIAL channel list\2.");

	} else {
		int remove_chan=0;

		if (*chan == '-') {
			*chan++;
			remove_chan=1;
		}

		if (!(ci = findchaninfo(chan))) {
			notice(s_ChanServ, source,
				"OFFICIAL ERROR: Channel '%s' is not registered.", chan);

		} else if (remove_chan && !(ci->flags & CS_OFFICIAL))
			notice(s_ChanServ, source,
				"OFFICIAL ERROR: Channel '%s' is not marked as an official channel.", chan);

		else if (!remove_chan && ci->flags & CS_OFFICIAL)
			notice(s_ChanServ, source,
				"OFFICIAL ERROR: Channel '%s' is already marked as an official channel.", chan);

		else if (remove_chan && ci->flags & CS_OFFICIAL) {
			ci->flags &= ~CS_OFFICIAL;
			official_list(ci, DEL);
			notice(s_ChanServ, source,
				"Channel '%s' is no longer marked as an official channel of %s.",
					chan, NETWORK_NAME);
			globops(NULL, "%s unset channel '%s' as an official channel of %s.",
				opertrace(u,0), chan, NETWORK_NAME);

			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);

		} else if (!remove_chan && !(ci->flags & CS_OFFICIAL)) {
			ci->flags |= CS_OFFICIAL;
			official_list(ci, ADD);
			notice(s_ChanServ, source,
				"Channel '%s' is now marked as an official channel of %s.",
					chan, NETWORK_NAME);
			globops(NULL, "%s set '%s' as an official channel of %s.",
				opertrace(u,0), chan, NETWORK_NAME);

			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);

		}


	}
}

/************************************************************************************/


static void do_op(const char *source)
{
	const char *chan = gettok(NULL, " ", CHANMAX);
	ChanInfo *ci;
	NickInfo *ni;
	char *onicklist[5], *tmp;
	User *u = finduser(source);
	int nicks=0, n=0, acc=0, i=0, dupd=0;


	if (!u)
		return;


	while ((tmp = strtok(NULL, " ")) != NULL)
	{
		dupd=0;
		for (i=0; i < nicks; i++) {
			if (sstrcmp(tmp, onicklist[i]) == 0) {
				dupd=1;
				break;
			}
		}
		if (dupd)
			continue;
		onicklist[nicks++] = tmp;
		if (nicks > 5)
			break;

	}

	if (!chan || !nicks) {
		notice(s_ChanServ, source,
			"Syntax: /%s OP <#channel> <nick1> [nick2] [nick3]...", s_ChanServ);
		notice(s_ChanServ, source,
			"\2/%s HELP OP\2 for more information.",
			s_ChanServ);

	} else if (!findchan(chan)) {
		notice(s_ChanServ, source,
			"OP ERROR: Channel '%s' currently is not in use! (empty)", chan);

	} else if (!(ci = findchaninfo(chan))) {
		notice(s_ChanServ, source,
			"OP ERROR: Channel '%s' is not registered!", chan);

	} else if (is_frozen(chan)) {
		notice(s_ChanServ, source,
			"OP ERROR: Channel '%s' is frozen! %s will not comply", chan, s_ChanServ);

	} else if ((acc=get_access(u, chan, NULL, 0)) < 3) {
		notice(s_ChanServ, source,
			"OP ERROR: You do not have access to use this command for channel %s.", chan);

	} else if ((sstrcmp(onicklist[n], source) == 0) && (nicks==1)) {
		if (acc==3) {
			if (is_voiced(source, chan))
				notice(s_ChanServ, source,
					"You have auto-voice access and are already voiced on channel %s!", chan);
			else {
				send_cmd(s_ChanServ, "MODE %s +v %s", chan, source);
				chan_addop(chan, source, 1);
			}
#ifdef HAS_HALFOP
		} else if (acc==4) {
			if (is_halfop(source, chan))
				notice(s_ChanServ, source,
					"You are already half-opped on channel %s!", chan);
			else {
				send_cmd(s_ChanServ, "MODE %s +h %s", chan, source);
				chan_addop(chan, source, 2);
			}
#endif
#ifdef UNREAL
		} else if (acc>=13) {
			int isop = is_chanop(source, chan);
			send_cmd(s_ChanServ, "MODE %s +%s%s %s %s", chan,
				isop ? "" : "o",
				acc==13 ? "a" : "q",
				!isop ? source : "", source);
			if (!isop)
				chan_addop(chan, source, 0);
			notice(s_ChanServ, source,
				"You have been set +%s%s on channel %s.",
				!isop ? "o" : "", acc==13 ? "a" : "q", chan);
#endif
		} else {
			if (is_chanop(source, chan))
				notice(s_ChanServ, source,
					"You are already opped on channel %s!", chan);
			else {
				send_cmd(s_ChanServ, "MODE %s +o %s", chan, source);
				chan_addop(chan, source, 0);
			}
		}


	} else if ((acc < 13) && (ci->flags & CS_OPGUARD)) {
		notice(s_ChanServ, source,
			"OP ERROR: Unless you are a founder of %s, you may not use CS OP on other nick's "
			"(except your own) when OPGUARD is set.", chan);

	} else {
		char buf[(NICKMAX * 6) + 6];
		char modebuf[7];
		*buf=0, *modebuf=0;
		for (i=0; i < nicks; i++) {

			if (!in_chan(onicklist[i], chan)) {
				notice(s_ChanServ, source,
					"Nick '%s' isn't in channel %s!", onicklist[i], chan);
				continue;
			}

			ni = findnick(onicklist[i]);

			if (ni && ni->flags & NS_NEVEROP) {
				notice(s_ChanServ, source,
					"Nick '%s' has NEVEROP setting enabled - not opping.", onicklist[i]);
				continue;
			}

			if (is_frozen(onicklist[i])) {
				notice(s_ChanServ, source,
					"Nick '%s' is frozen by the network.", onicklist[i]);
				continue;
			}


			acc=get_access(finduser(onicklist[i]), chan, NULL, 0);

			if ((acc==3) && is_voiced(onicklist[i], chan)) {
				notice(s_ChanServ, source,
					"Nick '%s' already voiced - not setting +v", onicklist[i]);
				continue;

#ifdef HAS_HALFOP
			} else if ((acc==4) && is_halfop(onicklist[i], chan)) {
				notice(s_ChanServ, source,
					"Nick '%s' already half-opped - not setting +h", onicklist[i]);
				continue;
#endif

			} else if (is_chanop(onicklist[i], chan)) {
				notice(s_ChanServ, source,
					"Nick '%s' already opped - not setting +o", onicklist[i]);
				continue;
			}

			if (acc < 3) {
				if (ci->flags & CS_OPGUARD)
					notice(s_ChanServ, onicklist[i],
					"You have been given op privileges on %s regardless of the OPGUARD setting. "
					"Please respect the founder's wishes.", chan);
			}
			if (*buf)
				strcat(buf, " ");
			strcat(buf, onicklist[i]);
			if (acc==3)
				strcat(modebuf, "v");
#ifdef HAS_HALFOP
			else if (acc==4)
				strcat(modebuf, "h");
#endif
			else
				strcat(modebuf, "o");

			chan_addop(chan, onicklist[i], 
				acc==3 ? 1 : acc==4 ? 2 : 0);
				
		}
		if (!*buf || !*modebuf) {
			notice(s_ChanServ, source,
				"OP ERROR: No valid nicks to be given operator status. "
				"Possible reasons may be listed above.");

		} else {
			send_cmd(s_ChanServ, "MODE %s +%s %s", chan, modebuf, buf);
			send_cmd(s_ChanServ, "NOTICE @+%s :%s!%s@%s used %s OP: %s",
				chan, source, u->username, u->host, s_ChanServ, buf);
		}

	}
}

/************************************************************************************/


static void do_remove(const char *source)
{
	ChanInfo *ci;
	User *u = finduser(source);
	const char *chan = gettok(NULL, " ", CHANMAX);
	const char *nick = gettok(NULL, " ", NICKMAX);
	char *ntrace;
	int acc=0;


	if (!u)
		return;

	if (chan && !nick)
		nick = source;


	if (!CHAN_REMOVE) {
		notice(s_ChanServ, source,
			"Sorry, the network has disabled the REMOVE command.");

	} else if (!chan) {
		notice(s_ChanServ, source,
			"Syntax: /%s REMOVE <#channel> [nick]", s_ChanServ);
		notice(s_ChanServ, source,
			"\2/%s HELP REMOVE\2 for more information.",
			s_ChanServ);

	} else if (!(ci = findchaninfo(chan))) {
		notice(s_ChanServ, source,
			"REMOVE ERROR: Channel '%s' is not registered!", chan);

	} else if (!nick_registered(nick)) {
		notice(s_ChanServ, source,
			"REMOVE ERROR: Nick '%s' is not registered!", nick);


	} else if (sstrcmp(nick, ci->founder) == 0) {
		notice(s_ChanServ, source,
			"REMOVE ERROR: This command may not be used on the founder's nick!");


	} else if (!is_identified(source, nick )) {
		notice(s_ChanServ, source,
			"REMOVE ERROR: You are not identified to the nick '%s'. "
			"Identify to %s and try again.", nick, s_NickServ);


	} else {

		if (sql_select("successor FROM chan WHERE id = %d", ci->id)) {
			row = fetch(res);
			if (row[0] && sstrcmp(row[0], nick) == 0) {
				notice(s_ChanServ, source,
				"You have been removed as successor for channel %s.", chan);
				send_sql("UPDATE chan SET successor=NULL WHERE id = %d", ci->id);
			}
		}

		if ((acc=get_access(u, chan, &ntrace, 0)) < 3) {
			notice(s_ChanServ, source,
			"You are not currently on any access list for channel %s.", chan);


		} else {
			if (update_caccess(ci, 0, nick, ntrace, NULL, 0, acc, 0) == 3) {
				notice(s_ChanServ, source,
				"You have been successfully removed from the access list of channel %s.",chan);
			} else {
				notice(s_ChanServ, source,
				"No entries found for nick '%s' on channel %s.", nick, chan);
			}
		}
	}
}

/************************************************************************************/

static void do_sendpass(const char *source)
{
	ChanInfo *ci;
	User *u = finduser(source);
	const char *chan = gettok(NULL, " ", CHANMAX);
	const char *email = strtok(NULL, " ");


#ifdef USE_ENCRYPTION
	notice(s_ChanServ, source,
		"GETPASS has been disabled by the network.");
	return;
#endif

	if (!u)
		return;

	else if (!AUTH_EMAIL) {
		notice(s_ChanServ, source,
			"E-Mail systems have been disabled by the network.");

	} else if (!chan || !email) {
		notice(s_ChanServ, source,
			"Syntax: /%s SENDPASS <#channel> <e-mail>", s_ChanServ);
		notice(s_ChanServ, source,
			"\2/%s HELP SENDPASS\2 for more information.", s_ChanServ);

	} else if (!(ci = findchaninfo(chan))) {
		notice(s_ChanServ, source,
			"SENDPASS ERROR: Channel %s is not registered!", chan);

	} else if (!sql_select("email,auth,flags FROM nick WHERE id = %d", ci->fid)) {
		notice(s_ChanServ, source,
			"SENDPASS ERROR: Could not retrieve e-mail information for channel %s.", chan);

	} else {
		row = fetch(res);

		if (!row[0]) {
			notice(s_ChanServ, source,
				"SENDPASS ERROR: Could not retrieve e-mail information for the founder "
				"of channel %s. An e-mail was never set at the time of registration.", chan);

		} else 	if (atol(row[1]) != 1) {
			notice(s_ChanServ, source,
				"SENDPASS ERROR: E-mail information was never fully authorized for the founder's "
				"nick of channel %s.", chan);

		} else if (atol(row[2]) & NS_MAILBLOCK) {
			notice(s_ChanServ, source,
				"SENDPASS ERROR: SENDPASS has been disabled for the founder's nick of %s.", chan);

		} else if (sstrcmp(row[0], email) != 0) {
			blog("CS_S", "[CS] S* %s (%s!%s@%s) [E-Mail discrepancy]",
				chan, u->nick, u->username, u->host);
			notice(s_ChanServ, source,
				"SENDPASS ERROR: The e-mail you specified does not match with the "
				"one on file!. This attempt has been logged.");
		} else {
			char emsg[512 + (CHANMAX*2) + (PASSMAX*2)];
			char esubj[60 + CHANMAX];

			if (!sql_select("pass FROM chan WHERE id = %d", ci->id)) {
				notice(s_ChanServ, source,
					"SENDPASS ERROR: Could not retrieve password information.");
				return;
			}
			row = fetch(res);

			if (check_email(chan, BLOCK_SENDPASS, 0)) {
				notice(s_ChanServ, source,
					"SENDPASS ERROR: A SENDPASS attempt has already been made recently. "
					"For security reasons, you may not attempt to SENDPASS this nick "
					"again for a disclosed period of time.");
				return;
			} else
				add_email(chan, BLOCK_SENDPASS);

			blog("CS_S", "[CS] S %s (%s!%s@%s)", chan, u->nick, u->username, u->host);
			globops(s_ChanServ, "%s used SENDPASS on %s", source, chan);
			notice(s_ChanServ, source,
				"Channel password has been sent via e-mail to %s.", email);

			snprintf(emsg, sizeof(emsg),
				"Someone, possibly you, has requested the password of channel %s "
				"to be sent to the founder's e-mail address. "
				"Your channel's password is '%s'.\n"
				"Log back onto %s and type - /%s IDENTIFY %s %s\n"
				"If you did not request this action to be carried out, you may "
				"wish to have one of the network services administrators investigate this.",
				chan, row[0], NETWORK_NAME, s_ChanServ, chan, row[0]);
			snprintf(esubj, sizeof(esubj), "%s password retrieval", NETWORK_NAME);

			e_mail(emsg, esubj, email, 1);
		}

	}


}

/************************************************************************************/

#define SET_PASS       0
#define SET_FOUNDER    1
#define SET_URL        2
#define SET_FLAGS      3
#define SET_SUCCESSOR  4

static void do_set(const char *source)
{
	ChanInfo *ci;
	const char *chan = gettok(NULL, " ", CHANMAX);
	const char *cmd = gettok(NULL, " ", CMDMAX);
	char *param = strtok(NULL, "");
	User *u = finduser(source);
	long showopt=0;

	if (!u)
		return;


	if (cmd && sstrcmp(cmd, "FOUNDER") == 0)
		param = (char *)source;


	if (!chan || !cmd || !param) {
		notice(s_ChanServ, source,
			"Syntax: /%s SET <#channel> <command> <parameters>", s_ChanServ);
		notice(s_ChanServ, source,
			"\2/%s HELP SET\2 for more information.",
			s_ChanServ);

		return;
	} else if (!(ci=findchaninfo(chan))) {
		notice(s_ChanServ, source,
			"SET ERROR: Channel \2%s\2 is not registered!", chan);
		return;


	} else if (is_frozen(chan)) {
		notice(s_ChanServ, source,
			"SET ERROR: Channel \2%s\2 is currently frozen!", chan);
		return;


	} else if (!is_founder(u,chan) && (!sstrcmp(cmd, "FOUNDER") || !sstrcmp(cmd, "SUCCESSOR")
		|| !sstrcmp(cmd, "PASSWORD"))) {
		notice(s_ChanServ, source,
			"SET ERROR: Permission denied, you are not identified as the founder to %s", chan);
		notice(s_ChanServ, source,
			"Identify to channel %s and try again.", chan);
		return;


	} else if (get_access(u, chan, NULL, 0) < 13) {
		notice(s_ChanServ, source,
			"SET ERROR: Permission denied, you are not recognized as a founder to %s", chan);
		return;


	} else if (ci->flags & CS_SUSPEND) {
		notice(s_ChanServ, source,
			"SET ERROR: Channel \2%s\2 is currently suspended.", chan);
		return;

	}

	if (sql_select("pass,founder,url,flags,successor FROM chan WHERE id = %d", ci->id))
		row = fetch(res);

	if (sstrcmp(cmd, "FOUNDER") == 0) {
		do_set_founder(u, ci, param);

	} else if (sstrcmp(cmd, "DESC") == 0) {
		do_set_desc(u, ci, param);

	} else if (sstrcmp(cmd, "IDENT") == 0) {
		showopt=do_set_ident(u, ci, param);

	} else if (sstrcmp(cmd, "KEEPTOPIC") == 0) {
		showopt=do_set_keeptopic(u, ci, param);

	} else if (sstrcmp(cmd, "LEAVEOPS") == 0) {
		showopt=do_set_leaveops(u, ci, param);

	} else if (sstrcmp(cmd, "MEMO") == 0) {
		do_set_memo(u, ci, param);

	} else if (sstrcmp(cmd, "MLOCK") == 0) {
		do_set_mlock(u, ci, param);

	} else if (sstrcmp(cmd, "NOTIFY") == 0) {
		showopt=do_set_notification(u, ci, param);

	} else if (sstrcmp(cmd, "OPGUARD") == 0) {
		showopt=do_set_opguard(u, ci, param);

	} else if (sstrcmp(cmd, "PASSWORD") == 0) {
		do_set_password(u, ci, param);

	} else if (sstrcmp(cmd, "PRIVATE") == 0) {
		showopt=do_set_private(u, ci, param);

	} else if (sstrcmp(cmd, "PROTECT") == 0) {
		showopt=do_set_protect(u, ci, param);

	} else if (sstrcmp(cmd, "RESTRICTED") == 0) {
		showopt=do_set_restricted(u, ci, param);

	} else if (sstrcmp(cmd, "SUCCESSOR") == 0) {
		do_set_successor(u, ci, param);

	} else if (sstrcmp(cmd, "TOPICLOCK") == 0) {
		do_set_topiclock(u, ci, param);
		showopt=atoi(row[SET_FLAGS]);

	} else if (sstrcmp(cmd, "TRIMBANS") == 0) {
		showopt=do_set_trimban(u, ci, param);

	} else if (sstrcmp(cmd, "UNSECURE") == 0) {
		showopt=do_set_unsecure(u, ci, param);

	} else if (sstrcmp(cmd, "URL") == 0) {
		do_set_url(u, ci, param);

	} else if (sstrcmp(cmd, "WELCOME") == 0) {
		do_set_welcome(u, ci, param);

	} else {
		notice(s_ChanServ, source, "Unknown SET option \2%s\2.", strupper((char *)cmd));
		notice(s_ChanServ, source, "For more information type "
			"\2/%s HELP SET\2", s_ChanServ);

	}

	if (showopt) {
		char buf[512];
		notice(s_ChanServ, source, "Options: %s", show_coptions(buf, showopt, chan));
	}
}

/************************************************************************************/

static void do_set_founder(User *u, ChanInfo *ci, char *param)
{
	if (!is_identified(u->nick, u->nick)) {
		notice(s_ChanServ, u->nick,
			"You must be using a registered nick and be identified before using the SET FOUNDER command.");

	} else {
		NickInfo *ni = findnick(u->nick);

		if (sstrcmp(row[SET_FOUNDER], u->nick) == 0) {
			notice(s_ChanServ, u->nick,
				"You are already recorded as the founder of channel %s", ci->name);
			return;

		} else if (MAX_CHANS && (check_maxchans(ni->id) >= MAX_CHANS)) {
			notice(s_ChanServ, u->nick, MAX_CHANS_REACHED, u->nick);
			return;
		}
		strncpy(ci->founder, u->nick, sizeof(ci->founder));
		ci->fid = ni->id;

		if (row[SET_SUCCESSOR] && sstrcmp(row[SET_SUCCESSOR], u->nick) == 0)
			send_sql("UPDATE chan SET successor=NULL where id = %d", ci->id);

		send_sql("UPDATE chan SET founder='%s',fid=%d WHERE id = %d",
			u->nick, ni->id, ci->id);
		notice(s_ChanServ, u->nick, "Under the nick %s, you are now the founder of %s.",
			u->nick, ci->name);

		blog("CS_O", "[CS] F %s (%s!%s@%s) [%s -> %s]",
			ci->name, u->nick, u->username, u->host, ci->founder, u->nick);

	}
}

/************************************************************************************/

static long do_set_ident(User *u, ChanInfo *ci, char *param)
{
	if (sstrcmp(param, "ON") == 0) {

		if (ci->flags & CS_IDENT) {
			notice(s_ChanServ, u->nick,
				"The \2Ident\2 option is already set for channel %s!", ci->name);
			return 0;
		} else {
			ci->flags |= CS_IDENT;
			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);
			notice(s_ChanServ, u->nick,
				"The \2Ident\2 option is now enabled for channel %s.", ci->name);
		}

	} else if (sstrcmp(param, "OFF") == 0) {


		if (!(ci->flags & CS_IDENT)) {
			notice(s_ChanServ, u->nick,
				"The \2Ident\2 option is already disabled for channel %s!", ci->name);
			return 0;
		} else {
			ci->flags &= ~CS_IDENT;
			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);
			notice(s_ChanServ, u->nick,
				"The \2Ident\2 option is now disabled for channel %s.", ci->name);
		}

	} else {
		notice(s_ChanServ, u->nick,
			"Syntax: /%s SET %s IDENT [ON|OFF]", s_ChanServ, ci->name);
		notice(s_ChanServ, u->nick,
			"\2/%s HELP SET IDENT\2 for more information.",
			s_ChanServ);
		return 0;
	}

	return ci->flags;
}

/************************************************************************************/

static long do_set_keeptopic(User *u, ChanInfo *ci, char *param)
{

	if (sstrcmp(param, "ON") == 0) {

		if (ci->flags & CS_KEEPTOPIC) {
			notice(s_ChanServ, u->nick,
				"The \2Keeptopic\2 option is already set for channel %s!", ci->name);
			return 0;
		} else {
			ci->flags |= CS_KEEPTOPIC;
			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);
			notice(s_ChanServ, u->nick,
				"ChanServ will now store the topic for channel %s.", ci->name);
		}

	} else if (sstrcmp(param, "OFF") == 0) {


		if (!(ci->flags & CS_KEEPTOPIC)) {
			notice(s_ChanServ, u->nick,
				"The \2Keeptopic\2 option is already disabled for channel %s!", ci->name);
			return 0;
		} else {
			ci->flags &= ~CS_KEEPTOPIC;
			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);
			notice(s_ChanServ, u->nick,
				"The \2Keeptopic\2 option is now disabled for channel %s.", ci->name);
		}

	} else {
		notice(s_ChanServ, u->nick,
			"Syntax: /%s SET %s KEEPTOPIC [ON|OFF]", s_ChanServ, ci->name);
		notice(s_ChanServ, u->nick,
			"\2/%s HELP SET KEEPTOPIC\2 for more information.",
			s_ChanServ);
		return 0;
	}

	return ci->flags;
}

/************************************************************************************/

static long do_set_leaveops(User *u, ChanInfo *ci, char *param)
{
	if (sstrcmp(param, "ON") == 0) {

		if (ci->flags & CS_LEAVEOPS) {
			notice(s_ChanServ, u->nick,
				"The \2Leave Ops\2 option is already set for channel %s!", ci->name);
			return 0;
		} else {
			ci->flags |= CS_LEAVEOPS;
			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);
			notice(s_ChanServ, u->nick,
				"When empty, the first user to join channel %s will now keep ops.", ci->name);
		}

	} else if (sstrcmp(param, "OFF") == 0) {


		if (!(ci->flags & CS_LEAVEOPS)) {
			notice(s_ChanServ, u->nick,
				"The \2Leave Ops\2 option is already disabled for channel %s!", ci->name);
			return 0;
		} else {
			ci->flags &= ~CS_LEAVEOPS;
			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);
			notice(s_ChanServ, u->nick,
				"The \2Leave Ops\2 option is now disabled for channel %s.", ci->name);
		}

	} else {
		notice(s_ChanServ, u->nick,
			"Syntax: /%s SET %s LEAVEOPS [ON|OFF]", s_ChanServ, ci->name);
		notice(s_ChanServ, u->nick,
			"\2/%s HELP SET LEAVEOPS\2 for more information.",
			s_ChanServ);
		return 0;
	}

	return ci->flags;
}

/************************************************************************************/


static void do_set_desc(User *u, ChanInfo *ci, char *param)
{
	send_sql("UPDATE chan SET descr = '%s' WHERE id = %d",
		param, ci->id);
	notice(s_ChanServ, u->nick, "Description for channel %s changed to: %s",
		ci->name, param);
}

/************************************************************************************/

static void do_set_memo(User *u, ChanInfo *ci, char *param)
{
	if (sstrcmp(param, "NONE") && sstrcmp(param, "AVOICE") &&
		sstrcmp(param, "AOP") && sstrcmp(param, "SOP") &&
		sstrcmp(param, "CF") && sstrcmp(param, "CFOUNDER") &&
		sstrcmp(param, "FOUNDER")) {

		notice(s_ChanServ, u->nick,
			"MEMO ERROR: Options may only be NONE, AVOICE, AOP, SOP, CFOUNDER or FOUNDER");
		return;

	} else if (sstrcmp(param, "NONE") == 0) {
		ci->mlevel = 0;
		notice(s_ChanServ, u->nick,
			"Channel memos for %s are now disabled.", ci->name);

	} else if (sstrcmp(param, "AVOICE") == 0) {
		ci->mlevel = 3;
		notice(s_ChanServ, u->nick,
			"Channel memo level for %s is now set to Auto-Voice.", ci->name);

	} else if (sstrcmp(param, "AOP") == 0) {
		ci->mlevel = 5;
		notice(s_ChanServ, u->nick,
			"Channel memo level for %s is now set to Auto-Op.", ci->name);

	} else if (sstrcmp(param, "SOP") == 0) {
		ci->mlevel = 10;
		notice(s_ChanServ, u->nick,
			"Channel memo level for %s is now set to Super-Op.", ci->name);

	} else if (sstrcmp(param, "FOUNDER") == 0) {
		ci->mlevel = 15;
		notice(s_ChanServ, u->nick,
			"Channel memo level for %s is now set to Founder only.", ci->name);

	} else {
		ci->mlevel = 13;
		notice(s_ChanServ, u->nick,
			"Channel memo level for %s is now set to Co-Founder.", ci->name);
	}

	send_sql("UPDATE chan SET mlevel = %d WHERE id = %d", ci->mlevel, ci->id);
}

/************************************************************************************/

static void do_set_mlock(User *u, ChanInfo *ci, char *param)
{
	Channel *c;
	const char *source = u->nick;
	char *mptr = param, modebuf[32], xmodes[64],  *end;
	const char *arg1 = strtok(param, " ");
	const char *arg2 = strtok(NULL, " ");
	const char *arg3 = strtok(NULL, " ");
	const char *chankey, *chanlimit;
	long modeson=0, modesoff=0;
	int set=-1, keyset=0, limitset=0;

	*xmodes=0;


	if (*mptr != '-' && *mptr != '+') {
		notice(s_ChanServ, source,
			"MLOCK error: Invalid mode parameters. (Must begin with either a '+' or '-').");
		return;
	}

	while (*mptr) {
		int dupd=0;
		char *valid = mptr;

		while (*valid++) {
			if (*valid == *mptr && *valid != '+' && *valid != '-') { dupd=1; break; }
		}


		if (dupd) {
			*mptr++;
			continue;
		}

		if (*mptr == '-') {
			set=0;
			*mptr++;
			continue;

		} else if (*mptr == '+') {
			set=1;
			*mptr++;
			continue;
		}
		switch (*mptr++) {

			case 'c':
				if (set)
					modeson |= CMODE_C;
				else
					modesoff |= CMODE_C;
				break;


			case 'i':
				if (set)
					modeson |= CMODE_I;
				else
					modesoff |= CMODE_I;
				break;


			case 'k':
				if (keyset)
					break;

				else if (set && !limitset && !arg2) {
					notice(s_ChanServ, u->nick, "mode 'k' ignored - no key given.");
					break;

				} else if (set && limitset && !arg3) {
					notice(s_ChanServ, u->nick, "mode 'k' ignored - no key given.");
					break;

				} else {
					keyset=1;

					if (set) {
						int i = strlen(xmodes);
						if (limitset)
							chankey = arg3;
						else
							chankey = arg2;

						if (strchr(chankey, '*')) {
							notice(s_ChanServ, u->nick,
								"mode 'k' ignored - invalid characters.");
							break;
						}

						modeson |= CMODE_K;
						if (*xmodes) {
							strcat(xmodes, " ");
							i++;
						}
						strncpy(xmodes+i, chankey, sizeof(xmodes)-i);
						send_sql("UPDATE chan SET chankey = '%s' WHERE id = %d",
							chankey, ci->id);
					} else {
						modesoff |= CMODE_K;
						send_sql("UPDATE chan SET chankey = NULL WHERE id = %d",
							ci->id);
					}

				}

				break;

			case 'l':
				if (limitset)
					break;

				else if (set && !keyset && !arg2) {
					notice(s_ChanServ, u->nick, "mode 'l' ignored - no limit given.");
					break;

				} else if (set && keyset && !arg3) {
					notice(s_ChanServ, u->nick, "mode 'l' ignored - no limit given.");
					break;

				} else {
					limitset=1;

					if (set) {
						int i = strlen(xmodes);
						if (keyset)
							chanlimit = arg3;
						else
							chanlimit = arg2;


						if ((strspn(chanlimit, "1234567890") != strlen(chanlimit)) ||
						(atoi(chanlimit) < 1) || (atoi(chanlimit) > 9999999)) {
							notice(s_ChanServ, u->nick,
								"mode 'l' ignored - invalid limit.");
							break;
						}

						modeson |= CMODE_L;
						if (*xmodes) {
							strcat(xmodes, " ");
							i++;
						}
						strncpy(xmodes+i, chanlimit, sizeof(xmodes)-i);
						send_sql("UPDATE chan SET chanlimit = %s WHERE id = %d",
							chanlimit, ci->id);
					} else {
						modesoff |= CMODE_L;
						send_sql("UPDATE chan SET chanlimit = NULL WHERE id = %d",
							ci->id);

					}

				}

				break;

			case 'm':
				if (set)
					modeson |= CMODE_M;
				else
					modesoff |= CMODE_M;
				break;


			case 'n':
				if (set)
					modeson |= CMODE_N;
				else
					modesoff |= CMODE_N;
				break;

			case 'p':
				if (set)
					modeson |= CMODE_P;
				else
					modesoff |= CMODE_P;
				break;

			case 's':
				if (set)
					modeson |= CMODE_S;
				else
					modesoff |= CMODE_S;
				break;


			case 't':
				if (set)
					modeson |= CMODE_T;
				else
					modesoff |= CMODE_T;
				break;


			case 'M':
				if (set)
					modeson |= CMODE_m;
				else
					modesoff |= CMODE_m;
				break;

			case 'O':
				if (set)
					modeson |= CMODE_o;
				else
					modesoff |= CMODE_o;
				break;

			case 'R':
				if (set)
					modeson |= CMODE_r;
				else
					modesoff |= CMODE_r;

				break;

		}
	}
	modeson |= CMODE_R;
	end = modebuf; *end=0;

	end += sprintf(end, "+r%s%s%s%s%s%s%s%s%s%s%s%s",
		modeson & CMODE_C ? "c" : "",
		modeson & CMODE_I ? "i" : "",
		modeson & CMODE_K ? "k" : "",
		modeson & CMODE_L ? "l" : "",
		modeson & CMODE_M ? "m" : "",
		modeson & CMODE_N ? "n" : "",
		modeson & CMODE_P ? "p" : "",
		modeson & CMODE_S ? "s" : "",
		modeson & CMODE_T ? "t" : "",
		modeson & CMODE_m ? "M" : "",
		modeson & CMODE_r ? "R" : "",
		modeson & CMODE_o ? "O" : "");

	if (modesoff)
		end += sprintf(end, "-%s%s%s%s%s%s%s%s%s%s%s%s",
			modesoff & CMODE_C ? "c" : "",
			modesoff & CMODE_I ? "i" : "",
			modesoff & CMODE_K ? "k" : "",
			modesoff & CMODE_L ? "l" : "",
			modesoff & CMODE_M ? "m" : "",
			modesoff & CMODE_N ? "n" : "",
			modesoff & CMODE_P ? "p" : "",
			modesoff & CMODE_S ? "s" : "",
			modesoff & CMODE_T ? "t" : "",
			modesoff & CMODE_m ? "M" : "",
			modesoff & CMODE_r ? "R" : "",
			modesoff & CMODE_o ? "O" : "");



	if (*modebuf)
		notice(s_ChanServ, source,
			"Mode lock changed to \2%s %s\2%son channel %s.",
			modebuf, *xmodes ? xmodes : "", *xmodes ? " " : "", ci->name);


	else
		notice(s_ChanServ, source,
			"Mode lock on channel %s has been reset to +r.", ci->name);

	send_sql("UPDATE chan SET modeson = '%d', modesoff = '%d'%s WHERE id = %d",
		modeson, modesoff, (!(modeson & CMODE_K)) ? ", chankey=NULL" : "", ci->id);



	if ((c = findchan(ci->name)))
		check_modes(c, ci);
}

/************************************************************************************/

static long do_set_notification(User *u, ChanInfo *ci, char *param)
{
	if (sstrcmp(param, "ON") == 0) {

		if (ci->flags & CS_NOTIFY) {
			notice(s_ChanServ, u->nick,
				"The \2Notify\2 option is already set for channel %s!", ci->name);
			return 0;
		} else {
			ci->flags |= CS_NOTIFY;
			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);
			notice(s_ChanServ, u->nick,
				"The \2Notify\2 option is now enabled for channel %s.", ci->name);
		}

	} else if (sstrcmp(param, "OFF") == 0) {


		if (!(ci->flags & CS_NOTIFY)) {
			notice(s_ChanServ, u->nick,
				"The \2Notify\2 option is already disabled for channel %s!", ci->name);
			return 0;
		} else {
			ci->flags &= ~CS_NOTIFY;
			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);
			notice(s_ChanServ, u->nick,
				"The \2Notify\2 option is now disabled for channel %s.", ci->name);
		}

	} else {
		notice(s_ChanServ, u->nick,
			"Syntax: /%s SET %s NOTIFY [ON|OFF]", s_ChanServ, ci->name);
		notice(s_ChanServ, u->nick,
			"\2/%s HELP SET NOTIFY\2 for more information.",
			s_ChanServ);
		return 0;
	}

	return ci->flags;
}

/************************************************************************************/

static long do_set_opguard(User *u, ChanInfo *ci, char *param)
{

	if (sstrcmp(param, "ON") == 0) {

		if (ci->flags & CS_OPGUARD) {
			notice(s_ChanServ, u->nick,
				"The \2Op Guard\2 option is already set for channel %s!", ci->name);
			return 0;
		} else {
			ci->flags |= CS_OPGUARD;
			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);
			notice(s_ChanServ, u->nick,
				"The \2Op Guard\2 option is now enabled for channel %s.", ci->name);
		}

	} else if (sstrcmp(param, "OFF") == 0) {


		if (!(ci->flags & CS_OPGUARD)) {
			notice(s_ChanServ, u->nick,
				"The \2Op Guard\2 option is already disabled for channel %s!", ci->name);
			return 0;
		} else {
			ci->flags &= ~CS_OPGUARD;
			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);
			notice(s_ChanServ, u->nick,
				"The \2Op Guard\2 option is now disabled for channel %s.", ci->name);
		}

	} else {
		notice(s_ChanServ, u->nick,
			"Syntax: /%s SET %s OPGUARD [ON|OFF]", s_ChanServ, ci->name);
		notice(s_ChanServ, u->nick,
			"\2/%s HELP SET OPGUARD\2 for more information.",
			s_ChanServ);
		return 0;
	}

	return ci->flags;
}

/************************************************************************************/

static void do_set_password(User *u, ChanInfo *ci, char *param)
{
	if (sstrcmp(row[SET_PASS], param) == 0) {
		notice(s_ChanServ, u->nick,
			"The password for channel %s is already set to %s!", ci->name, param);

	} else if (match_wild_nocase(param, ci->name) || strlen(param) < 5 ||
		!sstrcmp(param, ci->name) || !sstrcmp(param, u->nick) || strchr(param, ' ')) {
		notice(s_ChanServ, u->nick,
			"PASSWORD FAILURE: Please try again with a more obscure password. "
			"Passwords should be atleast five characters in length, should not "
			"be something easily guessed (ie: your real name or nick), and cannot "
			"contain space or tab characters. "
			"\2/%s HELP REGISTER\2 for more information.", s_ChanServ);

	} else if (strlen(param) > PASSMAX) {
		notice(s_ChanServ, u->nick,
			"PASSWORD FAILURE: Passwords may not exceed %d characters in length!", PASSMAX);


	} else if (check_escape_codes(param)) {
		notice(s_ChanServ, u->nick,
			"PASSWORD FAILURE: Your password has invalid characters. Please try again with "
			"common characters and don't use special colour/attribute codes.");

	} else {
		blog("CS_O", "[CS] P %s (%s!%s@%s)",
			ci->name, u->nick, u->username, u->host);
		send_sql("UPDATE chan SET pass = '%s' WHERE id = %d", param, ci->id);
		notice(s_ChanServ, u->nick,
			"Password for channel %s changed to: \2%s\2.", ci->name, param);
	}
}

/************************************************************************************/

static long do_set_private(User *u, ChanInfo *ci, char *param)
{
	if (sstrcmp(param, "ON") == 0) {

		if (ci->flags & CS_PRIVATE) {
			notice(s_ChanServ, u->nick,
				"The \2Private\2 option is already set for channel %s!", ci->name);
			return 0;
		} else {
			ci->flags |= CS_PRIVATE;
			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);
			notice(s_ChanServ, u->nick,
				"Channel %s will no longer be shown in channel listings.", ci->name);
		}

	} else if (sstrcmp(param, "OFF") == 0) {


		if (!(ci->flags & CS_PRIVATE)) {
			notice(s_ChanServ, u->nick,
				"The \2Private\2 option is already disabled for channel %s!", ci->name);
			return 0;
		} else {
			ci->flags &= ~CS_PRIVATE;
			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);
			notice(s_ChanServ, u->nick,
				"The \2Private\2 option is now disabled for channel %s.", ci->name);
		}

	} else {
		notice(s_ChanServ, u->nick,
			"Syntax: /%s SET %s PRIVATE [ON|OFF]", s_ChanServ, ci->name);
		notice(s_ChanServ, u->nick,
			"\2/%s HELP SET PRIVATE\2 for more information.",
			s_ChanServ);
		return 0;
	}

	return ci->flags;
}

/************************************************************************************/

static long do_set_protect(User *u, ChanInfo *ci, char *param)
{
	if (sstrcmp(param, "ON") == 0) {

		if (ci->flags & CS_PROTECT) {
			notice(s_ChanServ, u->nick,
				"The \2Protect\2 option is already set for channel %s!", ci->name);
			return 0;
		} else {
			ci->flags |= CS_PROTECT;
			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);
			notice(s_ChanServ, u->nick,
				"ChanServ will now protect the ops of channel %s.", ci->name);
		}

	} else if (sstrcmp(param, "OFF") == 0) {


		if (!(ci->flags & CS_PROTECT)) {
			notice(s_ChanServ, u->nick,
				"The \2Protect\2 option is already disabled for channel %s!", ci->name);
			return 0;
		} else {
			ci->flags &= ~CS_PROTECT;
			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);
			notice(s_ChanServ, u->nick,
				"The \2Protect\2 option is now disabled for channel %s.", ci->name);
		}

	} else {
		notice(s_ChanServ, u->nick,
			"Syntax: /%s SET %s PROTECT [ON|OFF]", s_ChanServ, ci->name);
		notice(s_ChanServ, u->nick,
			"\2/%s HELP SET PROTECT\2 for more information.",
			s_ChanServ);
		return 0;
	}

	return ci->flags;
}

/************************************************************************************/

static long do_set_restricted(User *u, ChanInfo *ci, char *param)
{
	if (sstrcmp(param, "ON") == 0) {

		if (ci->flags & CS_RESTRICTED) {
			notice(s_ChanServ, u->nick,
				"The \2Restricted\2 option is already set for channel %s!", ci->name);
			return 0;
		} else {
			ci->flags |= CS_RESTRICTED;
			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);
			notice(s_ChanServ, u->nick,
				"The \2Restricted\2 option is now enabled for channel %s.", ci->name);
		}

	} else if (sstrcmp(param, "OFF") == 0) {


		if (!(ci->flags & CS_RESTRICTED)) {
			notice(s_ChanServ, u->nick,
				"The \2Restricted\2 option is already disabled for channel %s!", ci->name);
			return 0;
		} else {
			ci->flags &= ~CS_RESTRICTED;
			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);
			notice(s_ChanServ, u->nick,
				"The \2Restricted\2 option is now disabled for channel %s.", ci->name);
		}

	} else {
		notice(s_ChanServ, u->nick,
			"Syntax: /%s SET %s RESTRICTED [ON|OFF]", s_ChanServ, ci->name);
		notice(s_ChanServ, u->nick,
			"\2/%s HELP SET RESTRICTED\2 for more information.",
			s_ChanServ);
		return 0;
	}

	return ci->flags;
}

/************************************************************************************/


static void do_set_successor(User *u, ChanInfo *ci, char *param)
{
	NickInfo *ni;

	if (!sstrcmp(param, "OFF") || !sstrcmp(param, "NONE")) {
		if (!row[SET_SUCCESSOR])
			notice(s_ChanServ, u->nick,
				"There is no successor set for channel %s!", ci->name);
		else {
			notice(s_ChanServ, u->nick,
				"Nick '%s' has been removed as successor for channel %s.",
				row[SET_SUCCESSOR], ci->name);
			blog("CS_O", "[CS] S %s (%s!%s@%s)",
				ci->name, u->nick, u->username, u->host);
			send_sql("UPDATE chan SET successor=NULL WHERE id = %d", ci->id);
		}


	} else if (sstrcmp(row[SET_FOUNDER], param) == 0) {
		notice(s_ChanServ, u->nick,
			"SUCCESSOR FAILURE: You may not set the successor the same as the current founder!");

	} else if (row[SET_SUCCESSOR] && sstrcmp(row[SET_SUCCESSOR], param) == 0) {
		notice(s_ChanServ, u->nick,
			"The successor for channel %s is already set to %s!", ci->name, param);

	} else if (!(ni = findnick(param))) {
		notice(s_ChanServ, u->nick,
			"SUCCESSOR FAILURE: Nick '%s' is not registered!", param);

	} else if (AUTH_EMAIL && !is_authed(param)) {
		notice(s_ChanServ, u->nick,
			"SUCCESSOR FAILURE: Nick '%s' has not authorized their nick!", param);

	} else if (ni->flags & NS_NOOP) {
		notice(s_ChanServ, u->nick,
			"SUCCESSOR FAILURE: Nick '%s' does not wish to be added to channel access lists!", param);


	} else {
		blog("CS_O", "[CS] S %s (%s!%s@%s)",
			ci->name, u->nick, u->username, u->host);
		send_sql("UPDATE chan SET successor = '%s' WHERE id = %d", param, ci->id);
		notice(s_ChanServ, u->nick,
			"Successor for channel %s updated to: \2%s\2.", ci->name, param);
	}
}

/************************************************************************************/

static void do_set_topiclock(User *u, ChanInfo *ci, char *param)
{

	if (!sstrcmp(param, "AVOICE") || !sstrcmp(param, "AOP")
		|| !sstrcmp(param, "SOP") || !sstrcmp(param, "CF")
		|| !sstrcmp(param, "CFOUNDER") || !sstrcmp(param, "FOUNDER")) {

		ci->topiclock =
			!sstrcmp(param, "AVOICE") ? 3 :
			!sstrcmp(param, "AOP") ? 5 :
			!sstrcmp(param, "SOP") ? 10 :
			!sstrcmp(param, "CF") ? 13 :
			!sstrcmp(param, "CFOUNDER") ? 13 :
			!sstrcmp(param, "FOUNDER") ? 15 : 0;

		notice(s_ChanServ, u->nick,
			"Topic lock for channel %s set to: \2%s\2.", ci->name, param);
		send_sql("UPDATE chan SET topiclock = %s WHERE id = %d",
			!sstrcmp(param, "AVOICE") ? "3" :
			!sstrcmp(param, "AOP") ? "5" :
			!sstrcmp(param, "SOP") ? "10" :
			!sstrcmp(param, "CF") ? "13" :
			!sstrcmp(param, "CFOUNDER") ? "13" :
			!sstrcmp(param, "FOUNDER") ? "15" : "", ci->id);

	} else if (!sstrcmp(param, "OFF")) {
		ci->topiclock=0;
		notice(s_ChanServ, u->nick,
			"Topic lock for channel %s has been disabled.", ci->name);
		send_sql("UPDATE chan SET topiclock = 0 WHERE id = %d", ci->id);
	} else {
		notice(s_ChanServ, u->nick,
		"Syntax: \2SET\2 %s TOPICLOCK [AVOICE|AOP|SOP|CFOUNDER|FOUNDER|OFF]", ci->name);
		notice(s_ChanServ, u->nick,
		"\2/%s HELP SET TOPICLOCK\2 for more information.", s_ChanServ);

	}
		
}

/************************************************************************************/

static long do_set_trimban(User *u, ChanInfo *ci, char *param)
{
	if (sstrcmp(param, "ON") == 0) {

		if (ci->flags & CS_TRIMBAN) {
			notice(s_ChanServ, u->nick,
				"The \2Trim Ban\2 option is already set for channel %s!", ci->name);
			return 0;
		} else {
			ci->flags |= CS_TRIMBAN;
			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);
			notice(s_ChanServ, u->nick,
				"The \2Trim Ban\2 option is now enabled for channel %s.", ci->name);
		}

	} else if (sstrcmp(param, "OFF") == 0) {


		if (!(ci->flags & CS_TRIMBAN)) {
			notice(s_ChanServ, u->nick,
				"The \2Trim Ban\2 option is already disabled for channel %s!", ci->name);
			return 0;
		} else {
			ci->flags &= ~CS_TRIMBAN;
			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);
			notice(s_ChanServ, u->nick,
				"The \2Trim Ban\2 option is now disabled for channel %s.", ci->name);
		}

	} else {
		notice(s_ChanServ, u->nick,
			"Syntax: /%s SET %s TRIMBAN [ON|OFF]", s_ChanServ, ci->name);
		notice(s_ChanServ, u->nick,
			"\2/%s HELP SET TRIMBAN\2 for more information.",
			s_ChanServ);
		return 0;
	}

	return ci->flags;
}


/************************************************************************************/

static long do_set_unsecure(User *u, ChanInfo *ci, char *param)
{
	if (sstrcmp(param, "ON") == 0) {

		if (ci->flags & CS_UNSECURE) {
			notice(s_ChanServ, u->nick,
				"The \2Unsecure\2 option is already set for channel %s!", ci->name);
			return 0;
		} else {
			ci->flags |= CS_UNSECURE;
			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);
			notice(s_ChanServ, u->nick,
				"The \2Unsecure\2 option is now enabled for channel %s. "
				"NOTE: Now anyone matching the founder's NickServ access list may "
				"make changes to the channel settings.", ci->name);
		}

	} else if (sstrcmp(param, "OFF") == 0) {


		if (!(ci->flags & CS_UNSECURE)) {
			notice(s_ChanServ, u->nick,
				"The \2Unsecure\2 option is already disabled for channel %s!", ci->name);
			return 0;
		} else {
			ci->flags &= ~CS_UNSECURE;
			send_sql("UPDATE chan SET flags = %d WHERE id = %d", ci->flags, ci->id);
			notice(s_ChanServ, u->nick,
				"The \2Unsecure\2 option is now disabled for channel %s.", ci->name);
		}

	} else {
		notice(s_ChanServ, u->nick,
			"Syntax: /%s SET %s UNSECURE [ON|OFF]", s_ChanServ, ci->name);
		notice(s_ChanServ, u->nick,
			"\2/%s HELP SET UNSECURE\2 for more information.",
			s_ChanServ);
		return 0;
	}

	return ci->flags;
}

/************************************************************************************/

static void do_set_url(User *u, ChanInfo *ci, char *param)
{
	int none = sstrcmp(param, "none");


	if (row[SET_URL] && sstrcmp(row[SET_URL], param) == 0) {
		notice(s_ChanServ, u->nick,
			"The URL for channel %s is already set to %s!", ci->name, param);

	} else if (!none || check_valid_url(u->nick, param)) {
		notice(s_ChanServ, u->nick,
			"URL for channel %s %s\2%s\2.", ci->name,
			none ? "has been changed to: " : "has been disabled",
			none ? param : "");


		if (none)
			send_sql("UPDATE chan SET url = '%s' WHERE id = %d",
				param, ci->id);
		else
			send_sql("UPDATE chan SET url = NULL WHERE id = %d",
				ci->id);

	}

}

/************************************************************************************/

static void do_set_welcome(User *u, ChanInfo *ci, char *param)
{
	if (!sstrcmp(param, "NONE") || !sstrcmp(param, "OFF")) {
		if (!ci->welcome) {
			notice(s_ChanServ, u->nick,
				"Welcome message for channel %s already disabled.", ci->name);
		} else {
			notice(s_ChanServ, u->nick,
				"Welcome message for channel %s has been disabled.", ci->name);
			free(ci->welcome);
			ci->welcome = NULL;
			send_sql("UPDATE chan SET welcome=NULL WHERE id=%d", ci->id);
		}

	} else if (strlen(param) > 250) {
		notice(s_ChanServ, u->nick,
			"ERROR: Welcome message too long in length.");
	} else {
		notice(s_ChanServ, u->nick,
			"Welcome message for channel %s %s to: %s",
			ci->name, ci->welcome ? "changed" : "set", param);

		send_sql("UPDATE chan SET welcome='%s' WHERE id=%d", param, ci->id);

		if (ci->welcome)
			free(ci->welcome);
		ci->welcome = sstrdup(param);
	}
}


/************************************************************************************/

static void do_unban(const char *source)
{
	Channel *c;
	const char *chan = gettok(NULL, " ", CHANMAX);
	const char *param = strtok(NULL, " ");

	User *u = finduser(source), *u2;

	if (!u)
		return;

	else if (!chan) {
		notice(s_ChanServ, source,
			"Syntax: /%s UNBAN <#channel> [<paramaters>]", s_ChanServ);
		notice(s_ChanServ, source,
			"\2/%s HELP UNBAN\2 for more information.",
			s_ChanServ);

	} else if (!(c=findchan(chan))) {
		notice(s_ChanServ, source,
			"UNBAN ERROR: Channel %s is not currently in use! (empty)", chan);

	} else if (!findchaninfo(chan)) {
		notice(s_ChanServ, source,
			"UNBAN ERROR: Channel %s is not registered!", chan);

	} else if (is_frozen(chan)) {
		notice(s_ChanServ, source,
			"UNBAN ERROR: Channel %s is currently frozen!", chan);


	} else if (get_access(u, chan, NULL, 0) < 3) {
		notice(s_ChanServ, source,
			"UNBAN ERROR: You do not have access to this command for channel %s.", chan);


	} else if (param && !(u2 = finduser(param))) {
		notice(s_ChanServ, source,
			"UNBAN ERROR: User '%s' is not online!", param);


#ifdef BAHAMUT
	} else {

		send_cmd(NULL, "SVSMODE %s -b %s", chan, param ? param : source);
		notice(s_ChanServ, source,
			"All banned masks have been lifted for nick '%s' on channel '%s'",
				param ? param : source, chan);
		if (param)
			send_cmd(s_ChanServ,
				"NOTICE @%s :%s used %s UNBAN '%s'.",
				chan,  source, s_ChanServ, param);
		else
			send_cmd(s_ChanServ,
				"NOTICE @%s :%s UNBAN has been used by '%s'.",
				chan,  s_ChanServ, source);

#else

	} else {
		struct ChanBans *cb, *cb2;
		int nbans=0;
		User *user;

		user = param ? u2 : u;

		cb = c->bans;
		while (cb) {
			cb2 = cb->next;
			if (match_host(user, cb->mask)) {
				send_cmd(s_ChanServ, "MODE %s -b %s", chan, cb->mask);
				chan_delban(chan, cb->mask);
				nbans++;
			}
			cb = cb2;
		}

		if (nbans) {

			if (param)
				send_cmd(s_ChanServ,
					"NOTICE @%s :%s used %s UNBAN '%s'.",
					chan,  source, s_ChanServ, param);
			else
				send_cmd(s_ChanServ,
					"NOTICE @%s :%s UNBAN has been used by '%s'.",
					chan,  s_ChanServ, source);


			notice(s_ChanServ, source,
				"%d banned masks were found matching %s userhost that have been "
				"lifted from channel %s", nbans, param ? "the" : "your", chan);
		} else
			notice(s_ChanServ, source,
				"No bans found matching %s userhost for channel %s.",
					param ? "the" : "your", chan);
#endif
	}
}

/************************************************************************************/

static void do_when(const char *source)
{
	ChanInfo *ci;
	const char *chan = gettok(NULL, " ", CHANMAX);
	const char *nick = gettok(NULL, " ", NICKMAX);

	if (!chan || !nick) {
		notice(s_ChanServ, source,
			"Syntax: /%s WHEN <#channel> <nick>", s_ChanServ);
		notice(s_ChanServ, source,
			"\2/%s HELP WHEN\2 for more information.",
			s_ChanServ);

	} else if (!(ci = findchaninfo(chan))) {
		notice(s_ChanServ, source,
			"WHEN ERROR: Channel '%s does is not registered!", chan);


	} else if ((get_access(finduser(source), chan, NULL, 0) < 3) && !is_oper(source)) {
		notice(s_ChanServ, source,
			"WHEN ERROR: You do not have access to this command for channel %s.", chan);

	} else {
		struct AccessList *a;

		for (a = ci->entries; a; a = a->next) {
			if (sstrcmp(a->entry, nick) == 0) {
				long addtime = (time(NULL) - a->when);

				if (addtime > 31536000)
					notice(s_ChanServ, source,
					"Entry '%s' was added by '%s' on %s (over %d year%s ago)",
					a->entry, a->instater, timebuf(a->when),
					addtime/31536000, (addtime/31536000 == 1) ? "" : "s");

				else if (addtime > 604800)
					notice(s_ChanServ, source,
					"Entry '%s' was added by '%s' on %s (over %d week%s ago)",
					a->entry, a->instater, timebuf(a->when),
					addtime/604800, (addtime/604800 == 1) ? "" : "s");

				else if (addtime > 86400)
					notice(s_ChanServ, source,
					"Entry '%s' was added by '%s' on %s (over %d day%s ago)",
					a->entry, a->instater, timebuf(a->when),
					addtime/86400, (addtime/86400 == 1) ? "" : "s");

				else if (addtime > 3600)
					notice(s_ChanServ, source,
					"Entry '%s' was added by '%s' on %s (over %d hour%s ago)",
					a->entry, a->instater, timebuf(a->when),
					addtime/3600, (addtime/3600 == 1) ? "" : "s");

				else
					notice(s_ChanServ, source,
					"Entry '%s' was added by '%s' on %s (under 1 hour ago)",
					a->entry, a->instater, timebuf(a->when));

				return;
			}
		}

		notice(s_ChanServ, source, "Entry '%s' not found on any access list for channel %s.",
			nick, chan);

	}
}

/************************************************************************************/

static void do_why(const char *source)
{
	const char *chan = gettok(NULL, " ", CHANMAX);
	const char *nick = gettok(NULL, " ", NICKMAX);
	ChanInfo *ci;
	User *u;
	char *ntrace=NULL;


	if (!chan || !nick) {
		notice(s_ChanServ, source,
			"Syntax: /%s WHY <#channel> <nick>", s_ChanServ);
		notice(s_ChanServ, source,
			"\2/%s HELP WHY\2 for more information.",
			s_ChanServ);

	} else if (!(ci=findchaninfo(chan))) {
		notice(s_ChanServ, source,
			"WHY ERROR: Channel '%s' is not registered!", chan);

	} else if (!(u = finduser(nick))) {
		notice(s_ChanServ, source,
			"WHY ERROR: Nick '%s' is not online!", nick);

	} else if ((get_access(finduser(source), chan, NULL, 0) < 3) && !is_oper(source)) {
		notice(s_ChanServ, source,
			"WHY ERROR: You do not have access to this command for channel %s.", chan);


	} else {
		int type=0, acc=0;

		if (is_founder(u, chan)) {
			notice(s_ChanServ, source,
				"Nick %s has founder access to channel %s. "
				"Reason: Identification to channel.", nick, chan);
			return;
		}

		acc = get_access(u, chan, &ntrace, 1);

		 if (!ntrace && acc) {
			notice(s_ChanServ, source,
				"Unknown WHY error. Please try again later.");
			return;
		}

		if (ntrace) {
			if (strchr(ntrace, '@'))
				type=1;
			else if (is_identified(nick, ntrace))
				type=2;
			else
				type=3;
		}

		switch (acc) {
			case -2:
			case -1:
				notice(s_ChanServ, source,
					"Nick %s does not have access to channel %s. "
					"Reason: Banned.%s%s%s", nick, chan,
					ntrace ? " (Ban Entry: " : "", ntrace ? ntrace : "", ntrace ? ")" : "");
				break;

			case 0:
				notice(s_ChanServ, source,
					"Nick %s does not have access to channel %s. "
					"Reason: Basic User.", nick, chan);
				break;


			case 13:
		 	case 10:
			case 5:
			case 3:
				notice(s_ChanServ, source,
					"Nick %s has %s access to channel %s. "
					"Reason: %s%s", nick,
					acc==13 ? "Co-Founder" :
					acc==10 ? "Super-Op" :
					acc==5  ? "Auto-Op" :
					acc==3  ? "Auto-Voice" : "Unknown", chan,
					type==1 ? "User's host matches channel's access list." :
					type==2 ? "Identification to nick " :
					type==3 ? "Recognition to nick " : "Uknown",
					type>1 ? ntrace : "");
				break;


			case 14:
			case 15:
				notice(s_ChanServ, source,
					"Nick %s has founder access to channel %s. "
					"Reason: %s (%s).", nick, chan,
					type==2 ? "Identification to the founder's nick" : 
						  "Recognition to the founder's nick", ntrace);

				break;

			default:
				notice(s_ChanServ, source,
					"Unknown WHY error. Please try again later.");
		}

	}

}

/************************************************************************************/

static void send_reminder(const char *chan, const char *email)
{
	ChanInfo *ci = findchaninfo(chan);
	char emsg[512 + CHANMAX];
	char esubj[40];

	if (!ci || ci->remind)
		return;

	ci->remind=1;


	snprintf(esubj, sizeof(esubj),
		"Your registration is about to expire");


	snprintf(emsg, sizeof(emsg),
		"Your channel '%s' on %s will expire in %d days. If you do not log on and use the "
		"the channel by then, your registration will be lost.",
		chan, NETWORK_NAME, REMINDER_DAYS);

	e_mail(emsg, esubj, email, 1);

}


/************************************************************************************/


void expire_chans()
{
	time_t expire = time(NULL) - (CHAN_EXPIRE*60*60*24);
	time_t near_expire = expire + (24*60*60*REMINDER_DAYS);
	unsigned long num_rows, i, regs = sql_select("name FROM chan");
	int fexpires=0, ftransfers=0;

	send_sql("DELETE FROM cflags WHERE held=0 AND marked=0 AND closed=0 AND frozen=0");


	num_rows=sql_select("founder,name,id FROM chan");

	for (i=0; i < num_rows; i++) {   /* founder expire / successor function */
		row = fetch(res);

		if (!nick_registered(row[0])) {
			if (succeed_chan(row[1]))
				ftransfers++;
			else {
				blog("CS_X", "[CS] X %s", row[1]);
				send_sql("DELETE FROM chan WHERE id = %d", atol(row[2]));
				send_sql("DELETE from cflags WHERE name = '%s'", row[1]);
				send_sql("DELETE from reason WHERE name = '%s'", row[1]);
				send_sql("DELETE FROM caccess WHERE chan = %d", atol(row[2]));
				del_ci_chan(row[1]);
				fexpires++;
			}
		}
	}


	num_rows=sql_select("name,id FROM chan WHERE lasttime <= %d AND flags < %d AND flags <> -1", expire, HELD_ID);


	for (i=0; i < num_rows; i++) {
		row = fetch(res);
		blog("CS_X", "[CS] X %s", row[0]);
		del_ci_chan(row[0]);
		send_sql("DELETE from cflags WHERE name = '%s'", row[0]);
		send_sql("DELETE from reason WHERE name = '%s'", row[0]);
		send_sql("DELETE FROM caccess WHERE chan = %d", atol(row[1]));
	}

	send_sql("DELETE FROM chan WHERE lasttime <= %d AND flags < %d AND flags <> -1", expire, HELD_ID);


	if (DISPLAY_UPDATES)
		globops(NULL, "Completed chan expire: (%ld/%ld) (%d successions)",
			(long)(mysql_affected_rows(&mysql)+fexpires), regs, ftransfers);

	num_rows=sql_select("chan.name,nick.email FROM chan,nick WHERE chan.lasttime <= %d "
		"AND chan.fid=nick.id AND chan.flags < %d AND chan.flags <> -1", near_expire, HELD_ID);

        for (i=0; i < num_rows; i++) {
                row = fetch(res);
                if (!row[1])
                        continue;
                send_reminder(row[0], row[1]);
        }

}


/************************************************************************************/

int succeed_chan(const char *chan)
{
	ChanInfo *ci = findchaninfo(chan);
	char emsg[512 + NICKMAX + PASSMAX + 32];
	char esubj[32 + NICKMAX];
	char buf[512];

	if (!ci)
		return 0;


	snprintf(buf, sizeof(buf), 
		"SELECT chan.successor,chan.pass,nick.email,nick.id FROM chan,nick "
		"WHERE chan.id = %ld AND chan.successor=nick.name", ci->id);


	if (!mysql_query(&mysql, buf)) {
		if ((res2 = mysql_store_result(&mysql)) == NULL)
			return 0;
		if (!mysql_num_rows(res2))
			return 0;
        } else
		return 0;

	row2 = fetch(res2);



	if (!row2[0] || !row2[1] || !row2[3])
		return 0;



	snprintf(esubj, sizeof(esubj),
		"%s registration update", NETWORK_NAME);

	snprintf(emsg, sizeof(emsg),
		"The founder's nick, %s, of the channel %s on %s has expired. "
		"You have been appointed the successor the channel. Congratulations on your new "
		"channel foundership. In case the previous founder did not inform you, the channel "
		"password is '%s'. It is highly recommended that you change this password to something else "
		"and keep it a secret.", ci->founder, ci->name, NETWORK_NAME, row2[1]);

	if (row2[2])
		e_mail(emsg, esubj, row2[2], 1);


	blog("CS_F", "CS F %s [%s -> %s] (Successor)", ci->name, ci->founder, row2[0]);

	strncpy(ci->founder, row2[0], sizeof(ci->founder));
	ci->fid = atol(row2[3]);
	send_sql("UPDATE chan SET successor=NULL,founder='%s',fid=%d WHERE id=%d",
		row2[0], atol(row2[3]), ci->id);


	mysql_free_result(res2);
	return 1;

}

/************************************************************************************/

/* When a channel akick is placed, or a channel is first CLOSED, this will boot any
   matching user currently in the channel. */

static void mass_check_kick(const char *chan)
{
	Channel *c = findchan(chan);
	struct ChanUsers *cu, *cu2;

	if (!c)
		return;

	cu = c->users;

	while (cu) {
		cu2 = cu->next;
		if (check_should_kick(cu->user, chan))
			chan_deluser(cu->user, c);
		cu = cu2;
	}

}

/************************************************************************************/

static void do_force_topic(const char *source)
{
	ChanInfo *ci;
	const char *chan = gettok(NULL, " ", CMDMAX);
	const char *topic = strtok(NULL, "");

	if (!chan || !topic) {
		notice(s_ChanServ, source,
			"Syntax: /%s TOPIC <#channel> <topic>", s_ChanServ);
		notice(s_ChanServ, source,
			"\2/%s HELP TOPIC\2 for more information.",
			s_ChanServ);
		return;

	} else if (strlen(topic) > 300) {
		notice(s_ChanServ, source,
			"Length of topic too long, please try again with an abbreviated topic.");
		return;
	}

	send_cmd(s_ChanServ, "TOPIC %s %s %lu :%s",
			chan, s_ChanServ, time(NULL), topic);

	if ((ci=findchaninfo(chan))) {
		if (ci->topic)
			free(ci->topic);

		ci->topic = sstrdup(topic);
		ci->topic_time = time(NULL);
		strncpy(ci->topic_setter, s_ChanServ, sizeof(ci->topic_setter));
		if (ci->flags & CS_KEEPTOPIC)
			send_sql("UPDATE chan SET topic = '%s', tsetter = '%s', ttime = %d WHERE id = %d",
				topic, s_ChanServ, time(NULL), ci->id);
	}
	notice(s_ChanServ, source,
		"Topic has been forcefully changed on channel %s", chan);
	globops(NULL, "%s has forcefully changed the topic of channel %s.",
		opertrace(finduser(source), 0), chan);

}


/************************************************************************************/
/************************************************************************************/

