/*
 * SirvNET Services are copyright (c) 1998-2006 Trevor Klingbeil.
 *    http://www.sirv.net  ::  ftp.sirv.net
 * This program is copyrighted software; see the file COPYING for details.
 */

#include "services.h"
#include "log.h"

#ifdef USE_SQL

MYSQL mysql;
MYSQL_RES *res, *res2, *res3;
MYSQL_ROW row, row2, row3;
MYSQL_FIELD *field;

static short sqldata=0;    /* Indicator to free result set before calling for more */

/*****************************************************************************/
/*****************************************************************************/


static int check_db_version();



/*****************************************************************************/

static inline char *escape(char *buf, const char *s)
{
	mysql_real_escape_string(&mysql, buf, s, strlen(s));
	return buf;
}

/*****************************************************************************/


/*
 * Sends a 'SELECT' command to the SQL server and automatically places the result
 * set in *res; If no data in result set, returns 0, otherwise returns number of rows.
 * The result set is then freed on the next call to sql_select()
 */


int sql_select(const char *fmt,...)
{
	va_list args;
	const char *ptr;
	char buf[4096], query[4096], pbuf[24], escaped[1025];
	char *s;
	int i=0;


	if (!sql_alive)
		return 0;

	if (sqldata) {
		mysql_free_result(res);
	}
	sqldata=0;

	*query = 0;
	va_start(args, fmt);

	for (ptr = fmt; *ptr != '\0'; ptr++) {
		if (*ptr != '%') {
			query[i++] = *ptr;
			continue;
		}
		*ptr++;
		switch(*ptr) {
			case 's':
				s = escape(escaped, va_arg(args, char *));
				while(*s)
					query[i++] = *s++;
				break;
			case 'd':
				sprintf(pbuf, "%d", va_arg(args, int));
				s = pbuf;
				while(*s)
					query[i++] = *s++;
				break;
			default:
				query[i++] = *ptr;
				break;
		}
	}

	va_end(args);
	query[i] = 0;

	snprintf(buf, sizeof(buf), "SELECT %s", query);

	if (mysql_query(&mysql, buf)) {
		if (debug)
			log("SQL query error! query = (%s) - error: %s",
			buf, mysql_error(&mysql));
		return 0;
	} else {
		if (debug_traffic)
			log("SQL query sent: %s", buf);
		if ((res = mysql_store_result(&mysql)) == NULL) {
			return 0;
		}
		sqldata=1;
		return mysql_num_rows(res);
	}
}

/*****************************************************************************/

int send_sql(const char *fmt,...)
{
	va_list args;
	const char *ptr;
	char query[4096], pbuf[24], escaped[1025];
	char *s;
	int i=0;


	if (!sql_alive)
		return 0;


	*query = 0;

	va_start(args, fmt);

	for (ptr = fmt; *ptr != '\0'; ptr++) {
		if (*ptr != '%') {
			query[i++] = *ptr;
			continue;
		}
		*ptr++;
		switch(*ptr) {
			case 's':
				s = escape(escaped, va_arg(args, char *));
				while(*s)
					query[i++] = *s++;
				break;
			case 'd':
				sprintf(pbuf, "%d", va_arg(args, int));
				s = pbuf;
				while(*s)
					query[i++] = *s++;
				break;
			default:
				query[i++] = *ptr;
				break;
		}
	}
	va_end(args);
	query[i] = 0;

	if (mysql_query(&mysql, query)) {
		if (debug)
			log("SQL query error! query = (%s) - error: %s",
			query, mysql_error(&mysql));
		return 0;
	} else if (debug_traffic)
			log("SQL query sent: %s", query);

	return 1;
}


/*****************************************************************************/

static inline void create_sirv_table()
{
	send_sql("CREATE TABLE sirv ("
		"version int(2), "
		"patch int(2),"
		"nickv int(2),"
		"chanv int(2),"
		"memov int(2),"
		"date int(32),"
		"log char(8),"
		"news int(8),"
		"webcheck char(128))");

	send_sql("INSERT into sirv VALUES ('%d','%d','%d','%d','%d','%d', NULL, 0, '%s')",
		DB_VER, DB_PAT, NICKVER, CHANVER, MEMOVER, time(NULL), WEB_VER);
}


static inline void create_nick_table()
{
	send_sql("CREATE TABLE nick ( "
		"id int(10) NOT NULL, "
		"name Char(%d) NOT NULL, "
		"pass Char(%d) NOT NULL, "
		"email Char(64), "
		"url char(128), "
		"icq char(20), "
		"usermask Char(128) NOT NULL, "
		"time int(32) NOT NULL, "
		"lasttime int(32) NOT NULL, "
		"flags int(10) NOT NULL, "
		"stamp int (8) NOT NULL, "
		"question Char(1),"
		"answer Char(32),"
		"auth int(2) NOT NULL, "
		"forward char(32) NOT NULL, "  /* not forwarding = 0 */
		"news int(6) NOT NULL, "
		"primary key(id))",
#ifdef USE_ENCRYPTION
		NICKMAX, 128);
#else
		NICKMAX, PASSMAX);
#endif

}


static inline void create_nickaccess_table()
{
	send_sql("CREATE TABLE naccess ( "
		"id int(10) NOT NULL, "
		"mask char(%d) NOT NULL)",
		HOSTMAX);

}

static inline void create_pending_table()
{
	send_sql("CREATE TABLE pending ( "
		"id int(10) NOT NULL, "
		"info Char(64) NOT NULL, "
		"authcode char(32) NOT NULL, "
		"time int(32) NOT NULL, "
		"type int(1) NOT NULL, "
		"primary key(id))");
}

static inline void create_flags_table()
{
	send_sql("CREATE TABLE flags ( "
		"name Char(%d) NOT NULL, "
		"held int(32) NOT NULL, "
		"marked int(32) NOT NULL, "
		"forbidden int(32) NOT NULL, "
		"frozen int(32) NOT NULL)",
		NICKMAX);

}

static inline void create_oper_table()
{
	send_sql("CREATE TABLE oper ( "
		"name Char(%d) NOT NULL, "
		"access int(1) NOT NULL, "
		"pass char(%d) NOT NULL, "
		"instater char(%d) NOT NULL, "
		"stamp int (8) NOT NULL, "
		"sendlog int(1) NOT NULL)",
		NICKMAX, NICKMAX, NICKMAX);
}

static inline void create_desc_table()
{
	send_sql("CREATE TABLE reason ( "
		"name Char(%d) NOT NULL, "
		"type Char(20) NOT NULL, "
		"reason Char(255) NOT NULL,"
		"instater Char(%d) NOT NULL)",
		NICKMAX, NICKMAX);
}

static inline void create_jupiter_table()
{
	send_sql("CREATE TABLE jupiter ( "
		"name Char(128) NOT NULL, "
		"reason Char(255) NOT NULL,"
		"instater Char(%d) NOT NULL)",
		NICKMAX);
}

static inline void create_akill_table()
{
	send_sql("CREATE TABLE akill ( "
		"id int(10) NOT NULL, "
		"mask Char(128) NOT NULL, "
		"time int(32) NOT NULL, "
		"expire int(32) NOT NULL, "
		"reason char(255) NOT NULL, "
		"instater char(%d) NOT NULL, "
		"ticket Char(30) NOT NULL, "
		"primary key(id))",
		NICKMAX);
}

static inline void create_trigger_table()
{
	send_sql("CREATE TABLE triggers ( "
		"mask Char(128) NOT NULL, "
		"number int(3) NOT NULL, "
		"reason Char(255) NOT NULL,"
		"instater Char(%d) NOT NULL)",
		NICKMAX);
}

static inline void create_sqline_table()
{
	send_sql("CREATE TABLE sqline ( "
		"name Char(%d) NOT NULL, "
		"reason Char(255) NOT NULL,"
		"instater Char(%d) NOT NULL)",
		NICKMAX, NICKMAX);
}

static inline void create_sgline_table()
{
	send_sql("CREATE TABLE sgline ( "
		"name Char(255) NOT NULL, "
		"reason Char(200) NOT NULL, "
		"instater Char(%d) NOT NULL)",
		NICKMAX);
}

static inline void create_zline_table()
{
	send_sql("CREATE TABLE zline ( "
		"ip Char(255) NOT NULL, "
		"reason Char(200) NOT NULL, "
		"instater Char(%d) NOT NULL)",
		NICKMAX);
}

static inline void create_ignore_table()
{
	send_sql("CREATE TABLE silence ( "
		"mask Char(128) NOT NULL, "
		"time int(32) NOT NULL, "
		"expire int(32) NOT NULL, "
		"reason Char(200) NOT NULL, "
		"instater Char(%d) NOT NULL)",
		NICKMAX);
}

static inline void create_chan_table()
{
	send_sql("CREATE TABLE chan ( "
		"id int(10) NOT NULL, "
		"name Char(%d) NOT NULL, "
		"pass Char(%d) NOT NULL, "
		"founder Char(%d) NOT NULL, "
		"fid int(10) NOT NULL, "
		"descr Char(255) NOT NULL, "
		"url Char(128), "
		"welcome Char(255), "
		"time int(32) NOT NULL, "
		"lasttime int(32) NOT NULL, "
		"chankey Char(23), "
		"chanlimit int(10), "
		"modeson int(10) NOT NULL, "
		"modesoff int(10) NOT NULL, "
		"flags int(10) NOT NULL, "
		"topiclock int(2) NOT NULL,"
		"successor char(%d),"
		"topic blob,"
		"tsetter Char(%d),"
		"ttime int(32),"
		"mlevel int(2) NOT NULL, "
		"primary key(id))",
		CHANMAX, PASSMAX, NICKMAX, NICKMAX, NICKMAX);
}

static inline void create_chanaccess_table()
{
	send_sql("CREATE TABLE caccess ( "
		"chan int(10) NOT NULL, "
		"entry char(%d) NOT NULL, "
		"level int(2) NOT NULL, "
		"instater char(%d) NOT NULL, "
		"time int(32) NOT NULL,"
		"reason char(255))",
		HOSTMAX, NICKMAX);

}

static inline void create_cflags_table()
{
	send_sql("CREATE TABLE cflags ( "
		"name Char(%d) NOT NULL, "
		"held int(32) NOT NULL, "
		"marked int(32) NOT NULL, "
		"closed int(32) NOT NULL, "
		"frozen int(32) NOT NULL, "
		"expiry int(32))",
		NICKMAX);

}

static inline void create_memo_table()
{
	send_sql("CREATE TABLE memos ( "
		"recipient char(%d) NOT NULL, "
		"sender char(%d) NOT NULL, "
		"chan char(%d), "
		"memo blob NOT NULL, "
		"time int(32) NOT NULL, "
		"new int(2) NOT NULL, "
		"del int(2) NOT NULL, "
		"receipt int(2) NOT NULL)",
		NICKMAX, NICKMAX, CHANMAX);
	
}

static inline void create_news_table()
{
	send_sql("CREATE TABLE news ( "
		"id int(6) NOT NULL, "
		"time int(32) NOT NULL)");
	
}

/*****************************************************************************/

void check_tables()
{
	my_ulonglong rows;

	res = mysql_list_tables(&mysql, NULL);
	rows = mysql_num_rows(res);


	if (!rows) {
		log("WARNING: No tables found in '%s' database! Attempting to "
		"create new database tables...", SQL_DB);

		create_sirv_table();
		create_nick_table();
		create_flags_table();
		create_nickaccess_table();
		create_pending_table();
		create_oper_table();
		create_desc_table();
		create_jupiter_table();
		create_akill_table();
		create_trigger_table();
		create_sqline_table();
		create_sgline_table();
		create_zline_table();
		create_ignore_table();
		create_chan_table();
                create_chanaccess_table();
		create_cflags_table();
		create_memo_table();
		create_news_table();


	} else {
		int i=0;
		while ((row = fetch(res))) {
			if (!strcmp(row[0], "nick") || !strcmp(row[0], "sirv"))
				i++;
		}
		if (i != 2) {
			fatal_error("SQL data tables are non-existent in '%s' database! "
			"This should never happen unless there has been tamporing or "
			"an unexplained data failure. You may have to delete the '%s' "
			"database and restart!", SQL_DB, SQL_DB);
		}
		if (debug)
			log("tables in SQL database appear functional.");
	}






}

/*****************************************************************************/

/* If services go down before the internal timers delete the entry this will take care of it. */

static inline void delete_temp_opers()
{
	char *temp = "TEMP %";
	send_sql("DELETE FROM oper WHERE pass LIKE '%s'", temp);
}

/*****************************************************************************/

void check_mlines()
{
	Masters *m;
	int num_rows=0;

	for (m = masterlist; m; m = m->next) {
		if (!(sql_select("access FROM oper WHERE name = '%s'", m->nick))) {
			if (debug)
				log("M:line nick '%s' not found in oper database - adding...", m->nick);
			send_sql("INSERT INTO oper VALUES ('%s',4,'%s','%s',0,0)",
				m->nick, "default sirv3", m->nick);
		} else {
			row = fetch(res);
			if (atoi(row[0]) != 4)
				send_sql("UPDATE oper SET access = 4 WHERE name = '%s'", m->nick);
		}
	}
	if ((num_rows=sql_select("name FROM oper WHERE access = 4"))) {
		int valid=0,i;
		for (i=0; i < num_rows; i++) {
			row = fetch(res);
			for (m = masterlist; m; m = m->next) {
				if (sstrcmp(m->nick, row[0]) == 0) {
					valid=1;
					break;
				}
			}
			if (!valid) {
				if (debug)
					log("Services master '%s' not found in m:line, removing from database.",
					row[0]);
				send_sql("DELETE FROM oper WHERE name = '%s'", row[0]);
			} else
				valid=0;
		}
	}
}

/*****************************************************************************/


int connect_sql()
{
	if (!mysql_init(&mysql))
		fatal_error("Unable to initialize MySQL! - out of memory?");

	if (!mysql_real_connect(&mysql, SQL_HOST, SQL_USER, SQL_PASS,
		NULL, SQL_PORT, NULL, 0)) {
		log("Unable to connect to MySQL Database! Error - %s\n",
			mysql_error(&mysql));
		mysql_close(&mysql);
		return 0;
	} else
		log("Connected to MySQL Database.");

	sql_alive=1;
	res = mysql_list_dbs(&mysql, SQL_DB);

	if (!(row = fetch(res))) {
		log("WARNING: Database '%s' not found on SQL database server! "
		"attempting to create new database...", SQL_DB);

		if (!send_sql("create database %s", SQL_DB))
			fatal_error("Failed to create new database '%s' on "
			"SQL server '%s'\n", SQL_DB, SQL_HOST);
		else
			log("Successfully created new database '%s' on SQL server.", SQL_DB);
	}
	if (mysql_select_db(&mysql, SQL_DB))
		fatal_error("Unable to switch active database to '%s' -%s",
		SQL_DB, mysql_error(&mysql));
	else if (debug)
		log("Successfuly opened '%s' database on SQL server");

	mysql_free_result(res);



	if (debug)
		log("Checking database tables...");
	check_tables();

	if (debug)
		log("Checking to see if database is up to date...");
	if (!check_db_version())
		fatal_error("Database version check failed");

	if (debug)
		log("Confirming M:line nicks are secured...");
	check_mlines();
	delete_temp_opers();

	return 1;

}

/*****************************************************************************/

int ping_sql()
{
	if (debug)
		log("Pinging SQL server...");
	if (mysql_ping(&mysql)) {
		log("Lost connection to SQL database server!");
		mysql_close(&mysql);
		return 0;
	} else if (debug)
		log("Got ping response from SQL database server");

	return 1;


}

/*****************************************************************************/

static void patch_1_1()
{

	int i, num_rows = sql_select("name,pass FROM oper");

	for (i=0; i < num_rows; i++) {
		char pass[PASSMAX];
		row = fetch(res);


		if (row[1] && sstrcmp(row[1], "default sirv3")) {
			encryptm(row[1], strlen(row[1]), pass, PASSMAX, 1);
			send_sql("UPDATE oper SET pass = '%s' WHERE name = '%s'",
				pass, row[0]);
		}
	}
	num_rows = sql_select("flags,id FROM nick");

	for (i=0; i < num_rows; i++) {
		long flags=0;
		row = fetch(res);
		flags = atol(row[0]);

		if (flags & NS_ENCRYPTED) {
			flags &= ~NS_ENCRYPTED;
			send_sql("UPDATE nick SET flags=%d WHERE id = %s", flags, row[1]);
		}
	}
}


static void patch_1_2()
{
	int i, num_rows = sql_select("flags,id FROM nick");

	for (i=0; i < num_rows; i++) {
		long oflags=0, nflags=0;
		row = fetch(res);
		oflags = atol(row[0]);

		if (oflags & NS_ENFORCE)
			nflags |= NS_ENFORCE;
		if (oflags & NS_SECURE)
			nflags |= NS_SECURE;
		if (oflags & NS_HIDE_EMAIL)
			nflags |= NS_HIDE_EMAIL;
		if (oflags & NS_NOOP)
			nflags |= NS_NOOP;
		if (oflags & NS_NEVEROP)
			nflags |= NS_NEVEROP;
		if (oflags & MS_NOMEMO)
			nflags |= MS_NOMEMO;
		if (oflags & MS_EMAILMEMOS)
			nflags |= MS_EMAILMEMOS;
		if (oflags & NS_ENFORCE_HELD)
			nflags |= NS_ENFORCE_HELD;
		if (oflags & NS_PRIVATE)
			nflags |= NS_PRIVATE;
		if (oflags & NS_MAILBLOCK)
			nflags |= NS_MAILBLOCK;
		if (oflags & MS_MEMO_LOGON)
			nflags |= MS_MEMO_LOGON;
		if (oflags & MS_MEMO_NEW)
			nflags |= MS_MEMO_NEW;
		if (oflags & MS_MEMO_RCPT)
			nflags |= MS_MEMO_RCPT;
		if (oflags & NS_ENCRYPTED)
			nflags |= NS_ENCRYPTED;
		if (oflags & NS_HELD)
			nflags |= NS_HELD;
		send_sql("UPDATE nick SET flags = %d WHERE id = %d", nflags, atoi(row[1]));
	}



	num_rows = sql_select("flags,id FROM chan");

	for (i=0; i < num_rows; i++) {
		long oflags=0, nflags=0;
		row = fetch(res);
		oflags = atol(row[0]);

		if (oflags & CS_IDENT)
			nflags |= CS_IDENT;
		if (oflags & CS_OPGUARD)
			nflags |= CS_OPGUARD;
		if (oflags & CS_KEEPTOPIC)
			nflags |= CS_KEEPTOPIC;
		if (oflags & CS_PROTECT)
			nflags |= CS_PROTECT;
		if (oflags & CS_UNSECURE)
			nflags |= CS_UNSECURE;
		if (oflags & CS_PRIVATE)
			nflags |= CS_PRIVATE;
		if (oflags & CS_RESTRICTED)
			nflags |= CS_RESTRICTED;
		if (oflags & CS_LEAVEOPS)
			nflags |= CS_LEAVEOPS;
		if (oflags & CS_NOTIFY)
			nflags |= CS_NOTIFY;
		if (oflags & CS_LEVELOPER)
			nflags |= CS_LEVELOPER;
		if (oflags & CS_LEVELSOP)
			nflags |= CS_LEVELSOP;
		if (oflags & CS_LEVELSA)
			nflags |= CS_LEVELSA;
		if (oflags & CS_LEVELSRA)
			nflags |= CS_LEVELSRA;
		if (oflags & CS_LEVELMASTER)
			nflags |= CS_LEVELMASTER;
		if (oflags & CS_SUSPEND)
			nflags |= CS_SUSPEND;
		if (oflags & CS_OFFICIAL)
			nflags |= CS_OFFICIAL;
		if (oflags & CS_TRIMBAN)
			nflags |= CS_TRIMBAN;
		if (oflags & CS_ENCRYPTED)
			nflags |= CS_ENCRYPTED;
		if (oflags & CS_HELD)
			nflags |= CS_HELD;

		send_sql("UPDATE chan SET flags = %d WHERE id = %d", nflags, atoi(row[1]));
	}

	send_sql("ALTER TABLE sirv ADD webcheck char(128)");

	send_sql("UPDATE sirv SET webcheck='%s'", WEB_VER);

}


/*****************************************************************************/

static int check_db_version()
{
	int num_rows;

	num_rows = sql_select("version,patch FROM sirv");

	if (num_rows) {
		if (num_rows > 1) {
			log("SQL database error: More than one version row found in database "
				"information table!");
		} else if (num_rows==1) {
			row = fetch(res);
			if (row[0] && row[1]) {
				if (atoi(row[0]) == DB_VER && atoi(row[1]) == DB_PAT)
					return 1;
				else {
					if (atoi(row[1]) < 1)
							patch_1_1();

					if (atoi(row[1]) < 2)
							patch_1_2();

					if (atoi(row[1]) < 3)
							send_sql("ALTER TABLE trigger RENAME TO triggers");

					send_sql("UPDATE sirv SET patch=%d", DB_PAT);
					return 1;
				}
			}
		}
	} else
		log("SQL database error: No data returned from database on version request!");
	return 0;
}

/*****************************************************************************/

void backup_routine()
{
	char buf[4096];

	if (DISPLAY_UPDATES)
		globops(NULL, "Database successfully backed up. Running again in %d hour%s.",
			BACKUP_UPDATE, BACKUP_UPDATE==1 ? "" : "s");

	log("Backing up SQL database to %s/backup/%s.sql", DATDEST, SQL_DB);

	if (chdir("./backup") < 0)
		system("mkdir ./backup");
	else
		chdir("..");

	send_sql("LOCK TABLES sirv read,nick read,naccess read,pending read,flags read,"
		 "oper read,reason read,jupiter read,akill read,triggers read,sqline read,sgline read,"
		 "zline read,silence read,chan read,caccess read,cflags read,memos read,news read");
	send_sql("FLUSH TABLES WITH READ LOCK");

	snprintf(buf, sizeof(buf),
		"%s/mysqldump --host=%s --port=%d --user=%s --password=%s --opt %s > %s/backup/%s.sql",
		SQL_PATH, SQL_HOST, SQL_PORT, SQL_USER, SQL_PASS, SQL_DB, DATDEST, SQL_DB);
	system(buf);
	send_sql("UNLOCK TABLES");
}


/*****************************************************************************/

inline void sql_quit()
{
	mysql_close(&mysql);
}

/*****************************************************************************/
/*****************************************************************************/

#endif  /* USE_SQL */

